#!/bin/sh

#---------------------------------------------------------------------
#
#  This script checks whether all CVS/Root and CVS/Repository within
# the DBOBJECTROOT repository are consistently defined with the 
# current $DBVCSROOT value.
#
#  o Whenever one of these files (recursively found) does not match 
# it is modified accordingly.
#
#  o We expect the object repository to be prefixed by "Objects"
#
#  o All modified files are backed up into a file suffixed by ".old"
#
#  Arguments : none
#  Output    : the list of modified files.
#
#---------------------------------------------------------------------

cd ${DBOBJECTSROOT}

for file in `find . -name Root -print`
do
  root=`cat $file`
  if test ! ${root} = ${DBCVSROOT} ; then
    echo changing $file
    echo $DBCVSROOT >${file}.new; \
     mv ${file} ${file}.old; mv ${file}.new ${file}
  fi
done

cvs_repo=`echo $DBCVSROOT | sed 's#//*$##'`

for file in `find . -name Repository -print`
do
  repository=`cat $file | sed 's#//*Objects.*##'`
  d=`cat $file | sed 's#.*/Objects#Objects#'`
  if test ! "${repository}" = ${cvs_repo} ; then
    echo changing $file
    echo ${cvs_repo}/${d} >${file}.new; \
     mv ${file} ${file}.old; mv ${file}.new ${file}
  fi
done
