#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Db.h>

DbType DbTIntNew ();

/*---------------------------------------------------------------------------*/
int* DbIntReferenceGetAddress (DbReference* reference)
/*---------------------------------------------------------------------------*/
{
  DbObject object;
  int* result;

  object = DbReferenceGetObject (reference);

  if (!DbTypeIsInt (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbIntReferenceGetAddress>");
      return (NULL);
    }

  result = (int*) DbIntObjectGetAddress (object);
  return (result);
}

/*---------------------------------------------------------------------------*/
int* DbIntObjectLoad (char* name)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbLoadObject (name);
  if (DbGetError () != DbErrorOk) return (NULL);

  if (!DbTypeIsInt (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbIntObjectLoad>");
      return (NULL);
    }

  return ((int*) DbIntObjectGetAddress (object));
}

/*---------------------------------------------------------------------------*/
int* DbIntObjectLoadByTag (char* name, char* tag)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbLoadObjectByTag (name, tag);
  if (DbGetError () != DbErrorOk) return (NULL);

  if (!DbTypeIsInt (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbIntObjectLoad>");
      return (NULL);
    }

  return ((int*) DbIntObjectGetAddress (object));
}

/*---------------------------------------------------------------------------*/
int* DbIntObjectLoadByDate (char* name, char* date)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbLoadObjectByDate (name, date);
  if (DbGetError () != DbErrorOk) return (NULL);

  if (!DbTypeIsInt (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbIntObjectLoad>");
      return (NULL);
    }

  return ((int*) DbIntObjectGetAddress (object));
}

/*---------------------------------------------------------------------------*/
DbVector* DbIntVectorNew ()
/*---------------------------------------------------------------------------*/
{
  return (DbVectorNew (DbTIntNew ()));
}

/*---------------------------------------------------------------------------*/
void DbIntVectorInit (DbVector* vector)
/*---------------------------------------------------------------------------*/
{
  DbVectorInit (vector, DbTIntNew ());
}

/*---------------------------------------------------------------------------*/
int* DbIntVectorGetArray (DbVector* vector)
/*---------------------------------------------------------------------------*/
{
  int* result;
  DbType type;

  type = DbVectorGetElementType (vector);
  if (!DbTypeIsInt (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbIntVectorGetArray>");
      return (NULL);
    }

  result = (int*) DbVectorGetAddress (vector, 0);

  return (result);
}

/*---------------------------------------------------------------------------*/
int DbIntVectorGetElement (DbVector* vector, int index)
/*---------------------------------------------------------------------------*/
{
  int* result;
  DbType type;

  type = DbVectorGetElementType (vector);
  if (!DbTypeIsInt (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbIntVectorGetElement>");
      return (0);
    }

  if (index >= DbVectorGetSize (vector))
    {
      DbVectorExtend (vector, index + 1);
    }

  result = (int*) DbVectorGetAddress (vector, index);

  return (*result);
}

/*---------------------------------------------------------------------------*/
void DbIntVectorSetElement (DbVector* vector, 
                              int index, 
                              int value)
/*---------------------------------------------------------------------------*/
{
  int* result;
  DbType type;

  type = DbVectorGetElementType (vector);
  if (!DbTypeIsInt (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbIntVectorSetElement>");
      return;
    }

  if (index >= DbVectorGetSize (vector))
    {
      DbVectorExtend (vector, index + 1);
    }

  result = (int*) DbVectorGetAddress (vector, index);

  *result = value;
}

/*---------------------------------------------------------------------------*/
DbMatrix* DbIntMatrixNew (int rows, int columns)
/*---------------------------------------------------------------------------*/
{
  return (DbMatrixNew (DbTIntNew (), rows, columns));
}

/*---------------------------------------------------------------------------*/
void DbIntMatrixInit (DbMatrix* matrix, int rows, int columns)
/*---------------------------------------------------------------------------*/
{
  DbMatrixInit (matrix, DbTIntNew (), rows, columns);
}

/*---------------------------------------------------------------------------*/
int* DbIntMatrixGetArray (DbMatrix* matrix)
/*---------------------------------------------------------------------------*/
{
  int* result;
  DbType type;

  if (matrix == NULL) return (NULL);

  type = DbMatrixGetElementType (matrix);
  if (!DbTypeIsInt (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbIntMatrixGetArray>");
      return (NULL);
    }

  result = (int*) DbMatrixGetAddress (matrix, 0, 0);

  return (result);
}

/*---------------------------------------------------------------------------*/
int DbIntMatrixGetElement (DbMatrix* matrix, int row, int column)
/*---------------------------------------------------------------------------*/
{
  int* result;
  DbType type;

  if (matrix == NULL) return (0);

  type = DbMatrixGetElementType (matrix);
  if (!DbTypeIsInt (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbIntMatrixGetElement>");
      return (0);
    }

  result = (int*) DbMatrixGetAddress (matrix, row, column);

  return (*result);
}

/*---------------------------------------------------------------------------*/
void DbIntMatrixSetElement (DbMatrix* matrix, 
                              int row, 
                              int column, 
                              int value)
/*---------------------------------------------------------------------------*/
{
  int* result;
  DbType type;

  if (matrix == NULL) return;

  type = DbMatrixGetElementType (matrix);
  if (!DbTypeIsInt (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbIntMatrixSetElement>");
      return;
    }

  result = (int*) DbMatrixGetAddress (matrix, row, column);

  *result = value;
}

