/*---------------------------------------------------------------------------*/
/*        Copyright (c) 1996 LAL Orsay, UPS-IN2P3-CNRS (France).             */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions        */
/* are met:                                                                  */
/* 1. Redistributions of source code must retain the above copyright         */
/*    notice, this list of conditions and the following disclaimer.          */
/* 2. Redistributions in binary form must reproduce the above copyright      */
/*    notice, this list of conditions and the following disclaimer in the    */
/*    documentation and/or other materials provided with the distribution.   */
/* 3. All advertising materials mentioning features or use of this software  */
/*    must display the following acknowledgement:                            */
/*      This product includes software developed by the Computer Application */
/*      Development Group at LAL Orsay (Laboratoire de l'Accelerateur        */
/*      Linaire - UPS-IN2P3-CNRS).                                           */
/* 4. Neither the name of the Institute nor of the Laboratory may be used    */
/*    to endorse or promote products derived from this software without      */
/*    specific prior written permission.                                     */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE LAL AND CONTRIBUTORS ``AS IS'' AND       */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE     */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR        */
/* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE LAL OR CONTRIBUTORS BE      */
/* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR       */
/* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF      */
/* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS  */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN   */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)   */
/* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF    */
/* THE POSSIBILITY OF SUCH DAMAGE.                                           */
/*---------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <CSet.h>
#include <DbDefs.h>
#include <DbLock.h>
#include <CmMessage.h>

/*-----------------------------------------------------------------*/
typedef struct _DbClientRec
{
  char* name;
  DbLockMode lock_mode;
} DbClientRec;

typedef struct _DbBranchRec
{
  char* name;
  CHasher locks;
} DbBranchRec;

typedef struct _DbLockRec
{
  char* object_name;
  DbClient client;
} DbLockRec;

typedef struct _DbTriggerRec
{
  DbTriggerType trigger_type;
  DbClient client;
} DbTriggerRec;

typedef struct _DbTriggerSetRec
{
  char* object_name;
  CLList triggers;
} DbTriggerSetRec;
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
static void DbClientDelete (DbClient client);
static void DbBranchDelete (DbBranch branch);
static void DbLockDelete (DbLock lock);
static void DbTriggerDelete (DbTrigger trigger);
static void DbTriggerSetDelete (DbTriggerSet trigger_set);
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
static CHasher Clients             = NULL;
static GCClass DbClientGCClass     = NULL;
static CHasher Branches            = NULL;
static GCClass DbBranchGCClass     = NULL;
static CHasher Locks               = NULL;
static GCClass DbLockGCClass       = NULL;
static CHasher TriggerSets         = NULL;
static GCClass DbTriggerSetGCClass = NULL;
static GCClass DbTriggerGCClass    = NULL;
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
/*                                                                 */
/*     Clients                                                     */
/*                                                                 */
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
void DbClientRestoreLocksAndTriggers ()
/*-----------------------------------------------------------------*/
{
  FILE* search;
  FILE* file;

  search = popen ("ls .locks/*.locks", "r");
  if (search != NULL)
    {
      char line[256];
      char client_name[256];
      DbClient client;

      while (fgets (line, sizeof (line), search) != NULL)
        {
          char* nl;
          char* dot;

          nl = strrchr (line, '\n');
          if (nl != NULL) *nl = 0;

          strcpy (client_name, &line[7]);
          dot = strrchr (client_name, '.');
          if (dot != NULL) *dot = 0;

          client = DbClientNew (client_name);
          if (CmConnectNew (client_name) != NULL)
            {
              file = fopen (line, "r");
              if (file != NULL)
                {
                  while (fgets (line, sizeof (line), file) != NULL)
                    {
                      char* object_name;

                      nl = strrchr (line, '\n');
                      if (nl != NULL) *nl = 0;

                      object_name = line;

                      if (DbFindLock (object_name) == NULL)
                        {
                          DbLockNew (object_name, client);
                        }
                    }
                  
                  fclose (file);
                }
            }
          else
            {
              char file_name[256];

              sprintf (file_name, ".locks/%s.locks", client_name);
              unlink (file_name);
            }
        }

      pclose (search);
    }

  search = popen ("ls .triggers/*.triggers", "r");
  if (search != NULL)
    {
      char line[256];
      char client_name[256];
      DbClient client;

      while (fgets (line, sizeof (line), search) != NULL)
        {
          char* nl;
          char* dot;

          nl = strrchr (line, '\n');
          if (nl != NULL) *nl = 0;

          strcpy (client_name, &line[10]);
          dot = strrchr (client_name, '.');
          if (dot != NULL) *dot = 0;

          client = DbClientNew (client_name);
          if (CmConnectNew (client_name) != NULL)
            {
              file = fopen (line, "r");
              if (file != NULL)
                {
                  while (fgets (line, sizeof (line), file) != NULL)
                    {
                      char object_name[256];
                      int trigger_value;
                      DbTriggerType trigger_type;

                      nl = strrchr (line, '\n');
                      if (nl != NULL) *nl = 0;

                      sscanf (line, "%s %d", object_name, &trigger_value);

                      switch (trigger_value)
                        {
                          case DbTriggerNone :
                            trigger_type = DbTriggerNone;
                            break;
                          case DbTriggerCommit :
                            trigger_type = DbTriggerCommit;
                            break;
                          case DbTriggerUnlock :
                            trigger_type = DbTriggerUnlock;
                            break;
                        }

                      DbTriggerAdd (object_name,
                                    trigger_type, 
                                    client);
                    }
                  
                  fclose (file);
                }
            }
          else
            {
              char file_name[256];

              sprintf (file_name, ".triggers/%s.triggers", client_name);
              unlink (file_name);
            }
        }

      pclose (search);
    }
}

/*-----------------------------------------------------------------*/
static void DbClientUpdateLocksFile (DbClient client)
/*-----------------------------------------------------------------*/
{
  FILE* file = NULL;
  char file_name[256];

  if (client == NULL) return;

  sprintf (file_name, ".locks/%s.locks", client->name);

  if (Locks != NULL)
    {
      CIterator it;

      it = CHasherGiveIterator (Locks);
      for (; !CIteratorFinished (it); CIteratorForth (it))
        {
          DbLock lock;
          
          lock = (DbLock) CIteratorItem (it);
          if (DbLockGetClient (lock) == client)
            {
              if (file == NULL)
                {
                  file = fopen (file_name, "w");

                  if (file == NULL) break;
                }

              fputs (DbLockGetName (lock), file);
	      fputs ("\n",file);
            }
        }
      CIteratorDelete (it);
    }

  if (file != NULL)
    {
      fclose (file);
    }
  else
    {
      unlink (file_name);
    }
}

/*-----------------------------------------------------------------*/
static void DbClientUpdateTriggersFile (DbClient client)
/*-----------------------------------------------------------------*/
{
  FILE* file = NULL;
  char file_name[256];

  sprintf (file_name, ".triggers/%s.triggers", client->name);

  if (TriggerSets != NULL)
    {
      CIterator it;

      it = CHasherGiveIterator (TriggerSets);
      for (; !CIteratorFinished (it); CIteratorForth (it))
        {
          DbTriggerSet trigger_set;
          CIterator t_it;
          
          trigger_set = (DbTriggerSet) CIteratorItem (it);
          if (trigger_set->triggers == NULL) continue;

          t_it = CLListGiveIterator (trigger_set->triggers);
          for (; !CIteratorFinished (t_it); CIteratorForth (t_it))
            {
              DbTrigger trigger = (DbTrigger) CIteratorItem (it);

              if (DbTriggerGetClient (trigger) == client)
                {
                  if (file == NULL)
                    {
                      file = fopen (file_name, "w");
                      
                      if (file == NULL) break;
                    }

                  fprintf (file, "%s %d\n", 
                           DbTriggerSetGetName (trigger_set),
                           DbTriggerGetTriggerType (trigger));
                }
            }
          CIteratorDelete (t_it);
        }
      CIteratorDelete (it);
    }

  if (file != NULL)
    {
      fclose (file);
    }
  else
    {
      unlink (file_name);
    }
}

/*-----------------------------------------------------------------*/
DbClient DbFindClient (char* name)
/*-----------------------------------------------------------------*/
{
  DbClient client;

  if (Clients == NULL) return (NULL);
  if (name == NULL) return (NULL);

  client = (DbClient) CHasherFind (Clients, name);

  return (client);
}

/*-----------------------------------------------------------------*/
DbClient DbClientNew (char* name)
/*-----------------------------------------------------------------*/
{
  DbClient client;

  client = DbFindClient (name);
  if (client != NULL) return (client);

  client = (DbClient) malloc (sizeof (DbClientRec));

  client->name = (char*) malloc (strlen (name) + 1);
  strcpy (client->name, name);
  client->lock_mode = DbLockReadOnly;

  if (DbClientGCClass == NULL)
    {
      GCReferenceAssign (DbClientGCClass,
                         GCClassNew ("DbClient", 
                                     (GCDeleteMethod) DbClientDelete));
    }

  GCObjectNew (client, DbClientGCClass);

  if (Clients == NULL)
    {
      GCReferenceAssign (Clients, 
                         CStrHasherNew (100, (CValuator) DbClientGetName));
    }

  CHasherAdd (Clients, client);

  return (client);
}

/*-----------------------------------------------------------------*/
void DbClientRemove (DbClient client)
/*-----------------------------------------------------------------*/
{
  char file_name[256];

  if (client == NULL) return;

  if (Locks != NULL)
    {
      CIterator it;
      int done;

      do
        {
          done = 1;

          it = CHasherGiveIterator (Locks);
          for (; !CIteratorFinished (it); CIteratorForth (it))
            {
              DbLock lock;
              
              lock = (DbLock) CIteratorItem (it);
              if (DbLockGetClient (lock) == client)
                {
                  done = 0;

                  DbLockRemove (lock);
                }
            }
          CIteratorDelete (it);
        } while (!done);
    }

  sprintf (file_name, ".locks/%s.locks", client->name);
  unlink (file_name);

  if (TriggerSets != NULL)
    {
      CIterator it;

      it = CHasherGiveIterator (TriggerSets);
      for (; !CIteratorFinished (it); CIteratorForth (it))
        {
          DbTriggerSet trigger_set;

          trigger_set = (DbTriggerSet) CIteratorItem (it);
          DbTriggerSetRemoveTriggers (trigger_set, client);
        }
      CIteratorDelete (it);
    }

  
  sprintf (file_name, ".triggers/%s.triggers", client->name);
  unlink (file_name);

  CHasherRemove (Clients, client);

  GCCleanup ();
}

/*-----------------------------------------------------------------*/
static void DbClientDelete (DbClient client)
/*-----------------------------------------------------------------*/
{
  if (client == NULL) return;

  if (client->name != NULL)
    {
      free (client->name);
      client->name = NULL;
    }

  client->lock_mode = DbLockReadOnly;

  free (client);
}

/*-----------------------------------------------------------------*/
void DbClientSetLockMode (DbClient client, DbLockMode lock_mode)
/*-----------------------------------------------------------------*/
{
  if (client == NULL) return;

  client->lock_mode = lock_mode;
}

/*-----------------------------------------------------------------*/
DbLockMode DbClientGetLockMode (DbClient client)
/*-----------------------------------------------------------------*/
{
  if (client == NULL) return (DbLockReadOnly);

  return (client->lock_mode);
}

/*-----------------------------------------------------------------*/
int DbClientRequestWriteAccess (DbClient client)
/*-----------------------------------------------------------------*/
{
  if (client == NULL) return (0);

  return ((client->lock_mode & DbLockReadWrite) != 0);
}

/*-----------------------------------------------------------------*/
int DbClientIsSame (DbClient client, char* name)
/*-----------------------------------------------------------------*/
{
  if (client == NULL) return (0);
  if (client->name == NULL) return (0);
  if (name == NULL) return (0);

  return (!strcmp (client->name, name));
}

/*-----------------------------------------------------------------*/
char* DbClientGetName (DbClient client)
/*-----------------------------------------------------------------*/
{
  if (client == NULL) return (NULL);
  return (client->name);
}

/*-----------------------------------------------------------------*/
/*                                                                 */
/*     Branches                                                    */
/*                                                                 */
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
DbBranch DbFindBranch (char* name)
/*-----------------------------------------------------------------*/
{
  DbBranch branch;

  if (Branches == NULL) return (NULL);
  if (name == NULL) return (NULL);

  branch = (DbBranch) CHasherFind (Branches, name);

  return (branch);
}

/*-----------------------------------------------------------------*/
DbBranch DbBranchNew (char* name)
/*-----------------------------------------------------------------*/
{
  DbBranch branch;

  branch = DbFindBranch (name);
  if (branch != NULL) return (branch);

  branch = (DbBranch) malloc (sizeof (DbBranchRec));

  branch->name = (char*) malloc (strlen (name) + 1);
  strcpy (branch->name, name);
  branch->locks = NULL;

  if (DbBranchGCClass == NULL)
    {
      GCReferenceAssign (DbBranchGCClass,
                         GCClassNew ("DbBranch", 
                                     (GCDeleteMethod) DbBranchDelete));
    }

  GCObjectNew (branch, DbBranchGCClass);

  if (Branches == NULL)
    {
      GCReferenceAssign (Branches, 
                         CStrHasherNew (100, (CValuator) DbBranchGetName));
    }

  CHasherAdd (Branches, branch);

  return (branch);
}

/*-----------------------------------------------------------------*/
void DbBranchRemove (DbBranch branch)
/*-----------------------------------------------------------------*/
{
  if (branch == NULL) return;

  if (branch->locks != NULL)
    {
      CHasherErase (branch->locks);
      GCReferenceClear (branch->locks);
    }

  CHasherRemove (Branches, branch);

  GCCleanup ();
}

/*-----------------------------------------------------------------*/
static void DbBranchDelete (DbBranch branch)
/*-----------------------------------------------------------------*/
{
  if (branch == NULL) return;

  if (branch->name != NULL)
    {
      free (branch->name);
      branch->name = NULL;
    }

  if (branch->locks != NULL)
    {
      CHasherErase (branch->locks);
      GCReferenceClear (branch->locks);
    }

  free (branch);
}

/*-----------------------------------------------------------------*/
int DbBranchIsSame (DbBranch branch, char* name)
/*-----------------------------------------------------------------*/
{
  if (branch == NULL) return (0);
  if (branch->name == NULL) return (0);
  if (name == NULL) return (0);

  return (!strcmp (branch->name, name));
}

/*-----------------------------------------------------------------*/
char* DbBranchGetName (DbBranch branch)
/*-----------------------------------------------------------------*/
{
  if (branch == NULL) return (NULL);
  return (branch->name);
}

/*-----------------------------------------------------------------*/
/*                                                                 */
/*     Locks                                                       */
/*                                                                 */
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
DbLock DbFindLock (char* object_name)
/*-----------------------------------------------------------------*/
{
  DbLock lock;

  if (Locks == NULL) return (NULL);
  if (object_name == NULL) return (NULL);

  lock = (DbLock) CHasherFind (Locks, object_name);

  return (lock);
}

/*-----------------------------------------------------------------*/
DbLock DbLockNew (char* object_name, DbClient client)
/*-----------------------------------------------------------------*/
{
  DbLock lock;

  if (object_name == NULL) return (NULL);

  lock = DbFindLock (object_name);
  if (lock != NULL) 
    {
      if (client != lock->client)
        {
          GCReferenceClear (lock->client);
          DbClientUpdateLocksFile (lock->client);
          GCReferenceAssignIn (lock->client, lock, client);
          DbClientUpdateLocksFile (client);
        }

      return (lock);
    }

  lock = (DbLock) malloc (sizeof (DbLockRec));

  lock->object_name = (char*) malloc (strlen (object_name) + 1);
  strcpy (lock->object_name, object_name);
  lock->client    = NULL;

  if (DbLockGCClass == NULL)
    {
      GCReferenceAssign (DbLockGCClass,
                         GCClassNew ("DbLock", 
                                     (GCDeleteMethod) DbLockDelete));
    }

  GCObjectNew (lock, DbLockGCClass);

  if (Locks == NULL)
    {
      GCReferenceAssign (Locks, 
                         CStrHasherNew (100, (CValuator) DbLockGetName));
    }

  CHasherAdd (Locks, lock);

  GCReferenceAssignIn (lock->client, lock, client);

  DbClientUpdateLocksFile (client);

  return (lock);
}

/*-----------------------------------------------------------------*/
void DbLockRemove (DbLock lock)
/*-----------------------------------------------------------------*/
{
  DbClient client = NULL;

  if (lock == NULL) return;

  GCReferenceAssign (client, lock->client);

  GCReferenceClear (lock->client);
  CHasherRemove (Locks, lock);

  DbClientUpdateLocksFile (client);
  GCReferenceClear (client);
}

/*-----------------------------------------------------------------*/
DbLock DbFindLockInBranch (char* branch_name, char* object_name)
/*-----------------------------------------------------------------*/
{
  DbLock lock;
  DbBranch branch;

  branch = DbFindBranch (branch_name);
  if (branch == NULL) return (NULL);

  if (branch->locks == NULL) return (NULL);
  if (object_name == NULL) return (NULL);

  lock = (DbLock) CHasherFind (branch->locks, object_name);

  return (lock);
}

/*-----------------------------------------------------------------*/
DbLock DbLockNewInBranch (char* branch_name, 
                          char* object_name, 
                          DbClient client)
/*-----------------------------------------------------------------*/
{
  DbLock lock;
  DbBranch branch;

  if (branch_name == NULL) return (NULL);
  if (object_name == NULL) return (NULL);

  branch = DbBranchNew (branch_name);

  lock = DbFindLockInBranch (branch_name, object_name);
  if (lock != NULL) 
    {
      GCReferenceAssignIn (lock->client, lock, client);

      return (lock);
    }

  lock = (DbLock) malloc (sizeof (DbLockRec));

  lock->object_name = (char*) malloc (strlen (object_name) + 1);
  strcpy (lock->object_name, object_name);
  lock->client    = NULL;

  if (DbLockGCClass == NULL)
    {
      GCReferenceAssign (DbLockGCClass,
                         GCClassNew ("DbLock", 
                                     (GCDeleteMethod) DbLockDelete));
    }

  GCObjectNew (lock, DbLockGCClass);

  if (branch->locks == NULL)
    {
      GCReferenceAssign (branch->locks, 
                         CStrHasherNew (100, (CValuator) DbLockGetName));
    }

  CHasherAdd (branch->locks, lock);

  GCReferenceAssignIn (lock->client, lock, client);

  return (lock);
}

/*-----------------------------------------------------------------*/
void DbLockRemoveFromBranch (char* branch_name, DbLock lock)
/*-----------------------------------------------------------------*/
{
  DbBranch branch;

  if (lock == NULL) return;

  branch = DbFindBranch (branch_name);
  if (branch == NULL) return;
  if (branch->locks == NULL) return;

  GCReferenceClear (lock->client);
  CHasherRemove (branch->locks, lock);
}

/*-----------------------------------------------------------------*/
static void DbLockDelete (DbLock lock)
/*-----------------------------------------------------------------*/
{
  if (lock == NULL) return;

  if (lock->object_name != NULL)
    {
      free (lock->object_name);
      lock->object_name = NULL;
    }

  GCReferenceClear (lock->client);

  free (lock);
}

/*-----------------------------------------------------------------*/
DbLockMode DbLockGetLockMode (DbLock lock)
/*-----------------------------------------------------------------*/
{
  if (lock == NULL) return (DbLockReadOnly);
  if (lock->client == NULL) return (DbLockReadOnly);

  return (DbClientGetLockMode (lock->client));
}

/*-----------------------------------------------------------------*/
void DbLockSetClient (DbLock lock, DbClient client)
/*-----------------------------------------------------------------*/
{
  if (lock == NULL) return;

  GCReferenceAssignIn (lock->client, lock, client);
}

/*-----------------------------------------------------------------*/
DbClient DbLockGetClient (DbLock lock)
/*-----------------------------------------------------------------*/
{
  if (lock == NULL) return (NULL);

  return (lock->client);
}

/*-----------------------------------------------------------------*/
int DbLockPermitsReadForWrite (DbLock lock, char* name)
/*-----------------------------------------------------------------*/
/*                                                                 */
/*   This function checks whether a given lock permits read        */
/*  access to the application "name" in read write mode.           */
/*                                                                 */
/*   This is only possible when the lock is dead (NULL or noclient)*/
/*  or when the lock mode id ReadOnly                              */
/*                                                                 */
/*-----------------------------------------------------------------*/
{
    /*
      Not a real lock.
      */
  if (lock == NULL) return (1);
  if (lock->client == NULL) return (1);

  if (DbClientIsSame (lock->client, name))
    {
      return (1);
    }
  else
    {
      return (0);
    }
}

/*-----------------------------------------------------------------*/
int DbLockPermitsWrite (DbLock lock, char* name)
/*-----------------------------------------------------------------*/
/*                                                                 */
/*   This function checks whether a given lock permits write       */
/*  access to the application "name"                               */
/*                                                                 */
/*-----------------------------------------------------------------*/
{
    /*
      Not a real lock.
      */
  if (lock == NULL) return (0);
  if (lock->client == NULL) return (0);

  if (DbClientIsSame (lock->client, name))
    {
        /*
          The lock is owned by NAME.

          We assume that as soon as a lock is set on an object,
          this is enough to permit write.
          The original detection of permission should be done
          at lock creation time.
          (This is enforced by the fact that LockRemove clears the
          reference to the client)
          */

      return (1);
    }
  else
    {
        /*
          The lock is owned by another client. 
          */
      
      return (0);
    }
}

/*-----------------------------------------------------------------*/
char* DbLockGetOwner (DbLock lock)
/*-----------------------------------------------------------------*/
{
  if (lock == NULL) return (NULL);

  return (DbClientGetName (lock->client));
}

/*-----------------------------------------------------------------*/
char* DbLockGetName (DbLock lock)
/*-----------------------------------------------------------------*/
{
  if (lock == NULL) return (NULL);
  return (lock->object_name);
}

/*-----------------------------------------------------------------*/
/*                                                                 */
/*     TriggerSets                                                 */
/*                                                                 */
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
void DbSendTriggers (char* object_name,
                     DbTriggerType trigger_type)
/*-----------------------------------------------------------------*/
{
  DbTriggerSet trigger_set;
  CIterator it;
  CmMessage message = NULL;

  trigger_set = DbFindTriggerSet (object_name);
  if (trigger_set == NULL) return;

  if (trigger_set->triggers == NULL) return;

  it = CLListGiveIterator (trigger_set->triggers);
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbTrigger trigger = (DbTrigger) CIteratorItem (it);

      if (trigger != NULL)
        {
          if (trigger->trigger_type == trigger_type)
            {
              char* client_name = DbTriggerGetOwner (trigger);

              if (client_name != NULL)
                {
                  if (message == NULL)
                    {
                      message = CmMessageNew ();

                      CmMessageSetType (message, "DbTrigger");
                      CmMessagePutInt (message, trigger_type);
                      CmMessagePutText (message, object_name);
                    }

                  CmMessageSend (message, client_name);
                }
            }
        }
    }
  CIteratorDelete (it);

  if (message != NULL)
    {
      CmMessageDelete (message);
    }
}

/*-----------------------------------------------------------------*/
int DbTriggerAdd (char* object_name,
                  DbTriggerType trigger_type, 
                  DbClient client)
/*-----------------------------------------------------------------*/
{
  DbTriggerSet trigger_set;
  DbTrigger trigger;

  trigger_set = DbTriggerSetNew (object_name);
  if (trigger_set == NULL) return (0);

  trigger = DbTriggerSetFindTrigger (trigger_set, trigger_type, client);
  if (trigger != NULL) return (1);

  trigger = DbTriggerNew (trigger_type, client);
  if (trigger == NULL) return (0);

  if (trigger_set->triggers == NULL)
    {
      GCReferenceAssignIn (trigger_set->triggers, trigger_set, 
                           CLListNew ());
    }

  CLListAdd (trigger_set->triggers, trigger);

  return (1);
}

/*-----------------------------------------------------------------*/
void DbTriggerRemove (char* object_name,
                      DbTriggerType trigger_type, 
                      DbClient client)
/*-----------------------------------------------------------------*/
{
  DbTriggerSet trigger_set;
  DbTrigger trigger;

  trigger_set = DbFindTriggerSet (object_name);
  if (trigger_set == NULL) return;

  if (trigger_set->triggers == NULL) return;

  trigger = DbTriggerSetFindTrigger (trigger_set, trigger_type, client);
  if (trigger == NULL) return;

  DbTriggerClear (trigger);
  CLListRemove (trigger_set->triggers, trigger);
}

/*-----------------------------------------------------------------*/
DbTrigger DbTriggerSetFindTrigger (DbTriggerSet trigger_set,
                                   DbTriggerType trigger_type,
                                   DbClient client)
/*-----------------------------------------------------------------*/
{
  CIterator it;
  DbTrigger result = NULL;

  if (trigger_set == NULL) return (NULL);
  if (trigger_set->triggers == NULL) return (NULL);

  it = CLListGiveIterator (trigger_set->triggers);
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbTrigger trigger = (DbTrigger) CIteratorItem (it);
      if (trigger != NULL)
        {
          if ((trigger->trigger_type == trigger_type) &&
              (trigger->client == client))
            {
              result = trigger;
              break;
            }
        }
    }
  CIteratorDelete (it);

  return (result);
}

/*-----------------------------------------------------------------*/
DbTriggerSet DbFindTriggerSet (char* object_name)
/*-----------------------------------------------------------------*/
{
  DbTriggerSet trigger_set;

  if (TriggerSets == NULL) return (NULL);
  if (object_name == NULL) return (NULL);

  trigger_set = (DbTriggerSet) CHasherFind (TriggerSets, object_name);

  return (trigger_set);
}

/*-----------------------------------------------------------------*/
DbTriggerSet DbTriggerSetNew (char* object_name)
/*-----------------------------------------------------------------*/
{
  DbTriggerSet trigger_set;

  if (object_name == NULL) return (NULL);

  trigger_set = DbFindTriggerSet (object_name);
  if (trigger_set != NULL) 
    {
      return (trigger_set);
    }

  trigger_set = (DbTriggerSet) malloc (sizeof (DbTriggerSetRec));

  trigger_set->object_name = (char*) malloc (strlen (object_name) + 1);
  strcpy (trigger_set->object_name, object_name);
  trigger_set->triggers = NULL;

  if (DbTriggerSetGCClass == NULL)
    {
      GCReferenceAssign (DbTriggerSetGCClass,
                         GCClassNew ("DbTriggerSet", 
                                     (GCDeleteMethod) DbTriggerSetDelete));
    }

  GCObjectNew (trigger_set, DbTriggerSetGCClass);

  if (TriggerSets == NULL)
    {
      GCReferenceAssign (TriggerSets, 
                         CStrHasherNew (100, (CValuator) DbTriggerSetGetName));
    }

  CHasherAdd (TriggerSets, trigger_set);

  return (trigger_set);
}

/*-----------------------------------------------------------------*/
void DbTriggerSetClear (DbTriggerSet trigger_set)
/*-----------------------------------------------------------------*/
{
  if (trigger_set == NULL) return;

  if (trigger_set->triggers != NULL)
    {
      CLListErase (trigger_set->triggers);
      GCReferenceClear (trigger_set->triggers);
    }
}

/*-----------------------------------------------------------------*/
static void DbTriggerSetDelete (DbTriggerSet trigger_set)
/*-----------------------------------------------------------------*/
{
  if (trigger_set == NULL) return;

  if (trigger_set->object_name != NULL)
    {
      free (trigger_set->object_name);
      trigger_set->object_name = NULL;
    }

  if (trigger_set->triggers != NULL)
    {
      CLListErase (trigger_set->triggers);
      GCReferenceClear (trigger_set->triggers);
    }

  free (trigger_set);
}

/*-----------------------------------------------------------------*/
char* DbTriggerSetGetName (DbTriggerSet trigger_set)
/*-----------------------------------------------------------------*/
{
  if (trigger_set == NULL) return (NULL);
  return (trigger_set->object_name);
}

/*-----------------------------------------------------------------*/
void DbTriggerSetRemoveTriggers (DbTriggerSet trigger_set, DbClient client)
/*-----------------------------------------------------------------*/
{
  CIterator it;
  int done;

  if (trigger_set == NULL) return;
  if (trigger_set->triggers == NULL) return;

  do
    {
      done = 1;

      it = CLListGiveIterator (trigger_set->triggers);

      for (CIteratorFirst (it); !CIteratorFinished (it);)
        {
          DbTrigger trigger = (DbTrigger) CIteratorItem (it);
          
          if (DbTriggerGetClient (trigger) == client)
            {
              DbTriggerClear (trigger);
              CLListRemove (trigger_set->triggers, trigger);
              done = 0;
              break;
            }
          else
            {
              CIteratorForth (it);
            }
        }
      CIteratorDelete (it);
    } while (!done);
}

/*-----------------------------------------------------------------*/
/*                                                                 */
/*     Triggers                                                       */
/*                                                                 */
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
DbTrigger DbTriggerNew (DbTriggerType trigger_type, DbClient client)
/*-----------------------------------------------------------------*/
{
  DbTrigger trigger;

  if (client == NULL) return (NULL);
  if (trigger_type == DbTriggerNone) return (NULL);

  trigger = (DbTrigger) malloc (sizeof (DbTriggerRec));

  trigger->trigger_type = trigger_type;
  trigger->client       = NULL;

  if (DbTriggerGCClass == NULL)
    {
      GCReferenceAssign (DbTriggerGCClass,
                         GCClassNew ("DbTrigger", 
                                     (GCDeleteMethod) DbTriggerDelete));
    }

  GCObjectNew (trigger, DbTriggerGCClass);

  GCReferenceAssignIn (trigger->client, trigger, client);

  return (trigger);
}

/*-----------------------------------------------------------------*/
void DbTriggerClear (DbTrigger trigger)
/*-----------------------------------------------------------------*/
{
  if (trigger == NULL) return;

  GCReferenceClear (trigger->client);
  trigger->trigger_type == DbTriggerNone;
}

/*-----------------------------------------------------------------*/
static void DbTriggerDelete (DbTrigger trigger)
/*-----------------------------------------------------------------*/
{
  if (trigger == NULL) return;

  GCReferenceClear (trigger->client);

  free (trigger);
}

/*-----------------------------------------------------------------*/
DbTriggerType DbTriggerGetTriggerType (DbTrigger trigger)
/*-----------------------------------------------------------------*/
{
  if (trigger == NULL) return (DbTriggerNone);
  if (trigger->client == NULL) return (DbTriggerNone);

  return (trigger->trigger_type);
}

/*-----------------------------------------------------------------*/
void DbTriggerSetClient (DbTrigger trigger, DbClient client)
/*-----------------------------------------------------------------*/
{
  if (trigger == NULL) return;

  GCReferenceAssignIn (trigger->client, trigger, client);
}

/*-----------------------------------------------------------------*/
DbClient DbTriggerGetClient (DbTrigger trigger)
/*-----------------------------------------------------------------*/
{
  if (trigger == NULL) return (NULL);

  return (trigger->client);
}

/*-----------------------------------------------------------------*/
char* DbTriggerGetOwner (DbTrigger trigger)
/*-----------------------------------------------------------------*/
{
  if (trigger == NULL) return (NULL);

  return (DbClientGetName (trigger->client));
}
