/*---------------------------------------------------------------------------*/
/*        Copyright (c) 1996 LAL Orsay, UPS-IN2P3-CNRS (France).             */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions        */
/* are met:                                                                  */
/* 1. Redistributions of source code must retain the above copyright         */
/*    notice, this list of conditions and the following disclaimer.          */
/* 2. Redistributions in binary form must reproduce the above copyright      */
/*    notice, this list of conditions and the following disclaimer in the    */
/*    documentation and/or other materials provided with the distribution.   */
/* 3. All advertising materials mentioning features or use of this software  */
/*    must display the following acknowledgement:                            */
/*      This product includes software developed by the Computer Application */
/*      Development Group at LAL Orsay (Laboratoire de l'Accelerateur        */
/*      Linaire - UPS-IN2P3-CNRS).                                           */
/* 4. Neither the name of the Institute nor of the Laboratory may be used    */
/*    to endorse or promote products derived from this software without      */
/*    specific prior written permission.                                     */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE LAL AND CONTRIBUTORS ``AS IS'' AND       */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE     */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR        */
/* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE LAL OR CONTRIBUTORS BE      */
/* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR       */
/* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF      */
/* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS  */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN   */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)   */
/* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF    */
/* THE POSSIBILITY OF SUCH DAMAGE.                                           */
/*---------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <CSet.h>
#include <Db.h>

/*---------------------------------------------------------------------------*/
static int ReferenceCount = 0;
static int ReferenceNameCount = 0;
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
DbReference* DbReferenceNew ()
/*---------------------------------------------------------------------------*/
{
  DbReference* reference = NULL;

  reference = (DbReference*) malloc (sizeof (DbReference));
  ReferenceCount++;

  DbReferenceInit (reference);

  return (reference);
}

/*---------------------------------------------------------------------------*/
void DbReferenceDelete (DbReference* reference)
/*---------------------------------------------------------------------------*/
{
  if (reference == NULL) return;

  DbReferenceErase (reference);

  free (reference);
  ReferenceCount--;
}

/*---------------------------------------------------------------------------*/
void DbReferenceInit (DbReference* reference)
/*---------------------------------------------------------------------------*/
{
  if (reference == NULL) return;

  reference->name          = NULL;
  reference->object_handle = NULL;
}

/*---------------------------------------------------------------------------*/
void DbReferenceErase (DbReference* reference)
/*---------------------------------------------------------------------------*/
{
  if (reference == NULL) return;

  if (reference->name != NULL)
    {
      free (reference->name);
      ReferenceNameCount--;

      reference->name = NULL;
    }

  if (reference->object_handle != NULL)
    {
      DbObjectHandleRemove (reference->object_handle);
      reference->object_handle = NULL;
    }
}

/*---------------------------------------------------------------------------*/
void DbReferenceSetName (DbReference* reference, char* name)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  if (reference == NULL) return;
  if (name == NULL) return;
  if (strlen (name) == 0) return;

  if (reference->name != NULL)
    {
      if (!strcmp (reference->name, name)) return;

      free (reference->name);
      ReferenceNameCount--;
      reference->name = NULL;
    }

  if (reference->object_handle != NULL)
    {
      DbHandleClear (reference->object_handle);
    }

  reference->name = (char*) malloc (strlen (name) + 1);
  ReferenceNameCount++;

  strcpy (reference->name, name);
}

/*---------------------------------------------------------------------------*/
void DbReferenceUnload (DbReference* reference)
/*---------------------------------------------------------------------------*/
{
  if (reference == NULL) return;

  if (reference->object_handle != NULL)
    {
      DbHandleClear (reference->object_handle);
    }
}

/*---------------------------------------------------------------------------*/
void DbReferenceLoad (DbReference* reference, char* name)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  if (reference == NULL) return;
  DbReferenceSetName (reference, name);

  object = DbLoadObject (name);
  if (object == NULL) return;

  if (reference->object_handle == NULL)
    {
      reference->object_handle = DbObjectHandleNew ();
    }

  DbHandleSet (reference->object_handle, object);
}

/*---------------------------------------------------------------------------*/
void DbReferenceLoadByTag (DbReference* reference, char* name, char* tag)
/*---------------------------------------------------------------------------*/
{
  fprintf (stderr, "Db> Sorry DbReferenceLoadByTag not yet implemented\n");
}

/*---------------------------------------------------------------------------*/
void DbReferenceLoadByDate (DbReference* reference, char* name, char* date)
/*---------------------------------------------------------------------------*/
{
  fprintf (stderr, "Db> Sorry DbReferenceLoadByDate not yet implemented\n");
}

/*---------------------------------------------------------------------------*/
void DbReferenceUpdate (DbReference* reference)
/*---------------------------------------------------------------------------*/
{
  DbObject object;
  char* name;

  if (reference == NULL) return;
  name = DbReferenceGetName (reference);
  if (name == NULL) return;

  object = DbLoadObject (name);
  if (object == NULL) return;

  if (reference->object_handle == NULL)
    {
      reference->object_handle = DbObjectHandleNew ();
    }

  DbHandleSet (reference->object_handle, object);
}

/*---------------------------------------------------------------------------*/
void DbReferenceUpdateByTag (DbReference* reference, char* name, char* tag)
/*---------------------------------------------------------------------------*/
{
  fprintf (stderr, "Db> Sorry DbReferenceUpdateByTag not yet implemented\n");
}

/*---------------------------------------------------------------------------*/
void DbReferenceUpdateByDate (DbReference* reference, char* name, char* date)
/*---------------------------------------------------------------------------*/
{
  fprintf (stderr, "Db> Sorry DbReferenceUpdateByDate not yet implemented\n");
}

/*---------------------------------------------------------------------------*/
void DbReferenceCommit (DbReference* reference)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  if (reference == NULL) return;

  object = DbReferenceGetObject (reference);
  if (object == NULL) return;

  DbObjectCommit (object);
}

/*---------------------------------------------------------------------------*/
DbObject DbReferenceGetObject (DbReference* reference)
/*---------------------------------------------------------------------------*/
{
  if (reference == NULL) return (NULL);
  if (reference->object_handle == NULL) return (NULL);

  return ((DbObject) DbHandleGet (reference->object_handle));
}

/*---------------------------------------------------------------------------*/
DbType DbReferenceGetType (DbReference* reference)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbReferenceGetObject (reference);

  return (DbObjectGetType (object));
}

/*---------------------------------------------------------------------------*/
char* DbReferenceGetName (DbReference* reference)
/*---------------------------------------------------------------------------*/
{
  if (reference == NULL) return (NULL);

  return (reference->name);
}
