/*---------------------------------------------------------------------------*/
/*        Copyright (c) 1996 LAL Orsay, UPS-IN2P3-CNRS (France).             */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions        */
/* are met:                                                                  */
/* 1. Redistributions of source code must retain the above copyright         */
/*    notice, this list of conditions and the following disclaimer.          */
/* 2. Redistributions in binary form must reproduce the above copyright      */
/*    notice, this list of conditions and the following disclaimer in the    */
/*    documentation and/or other materials provided with the distribution.   */
/* 3. All advertising materials mentioning features or use of this software  */
/*    must display the following acknowledgement:                            */
/*      This product includes software developed by the Computer Application */
/*      Development Group at LAL Orsay (Laboratoire de l'Accelerateur        */
/*      Linaire - UPS-IN2P3-CNRS).                                           */
/* 4. Neither the name of the Institute nor of the Laboratory may be used    */
/*    to endorse or promote products derived from this software without      */
/*    specific prior written permission.                                     */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE LAL AND CONTRIBUTORS ``AS IS'' AND       */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE     */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR        */
/* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE LAL OR CONTRIBUTORS BE      */
/* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR       */
/* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF      */
/* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS  */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN   */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)   */
/* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF    */
/* THE POSSIBILITY OF SUCH DAMAGE.                                           */
/*---------------------------------------------------------------------------*/

/* 
        DbType.c
         Created           : 27-APR-1992 by Christian Arnault
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <CSet.h>
 
#include <DbDefs.h>
#include <DbType.h>
#include <DbString.h>
#include <DbReference.h>
#include <DbVector.h>
#include <DbMatrix.h>

/*---------------------------------------------------------------------*/
#define DbBuildAlignStruct(type) static struct {\
  char prev;\
  type me;\
  int next; }* s = 0

#define DbComputeAlign ((int) (((char*) &s->me) - ((char*) &s->prev)))
/*---------------------------------------------------------------------*/

/*-- Type definitions. ------------------------------------------------*/
typedef unsigned int hexa;

typedef enum
{
  read_only,
  write_only,
  visit_only
} IteratorMode;

typedef void  (*DbRoutine) (DbType This);
typedef void* (*DbPtrRoutine) (DbType This);
typedef int   (*DbIntRoutine) (DbType This);
typedef char* (*DbStringRoutine) (DbType This);
typedef char* (*DbOutputRoutine) (DbType This, void* address);
typedef DbTypeIterator (*DbIteratorRoutine) (DbType This, 
                                             IteratorMode mode,
                                             void* address);
typedef DbType (*DbTypeRoutine) (DbType This);
typedef void   (*DbInitRoutine) (DbType This, void* address);
typedef int    (*DbCompareRoutine) (DbType This, DbType other);
typedef DbType (*DbUsedStructureRoutine) (DbType This);

typedef struct _DbTypeClassRec* DbTypeClass;
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
typedef struct _DbTypeClassRec 
{
  char* ClassName;

  DbRoutine Clear;
  DbStringRoutine Name;
  DbStringRoutine CName;
  DbStringRoutine CppName;
  DbStringRoutine SubTypeName;
  DbIntRoutine Size;
  DbIntRoutine Alignment;
  DbPtrRoutine Instanciate;
  DbInitRoutine Init;
  DbInitRoutine Erase;
  DbIteratorRoutine GiveIterator;
  DbOutputRoutine Output;
  DbCompareRoutine IsSame;
  DbUsedStructureRoutine UsedStructure;
} DbTypeClassRec;

typedef struct _DbTypeRec 
{
  DbTypeClass classDef;
} DbTypeRec;
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
typedef void (*DbIRoutine) (DbTypeIterator This);
typedef void (*DbIMoveRoutine) (DbTypeIterator This, char* key);
typedef DbTypeIterator (*DbIteratorIRoutine) (DbTypeIterator This);
typedef char* (*DbIKeyRoutine) (DbTypeIterator This);

typedef struct _DbTypeIteratorClassRec* DbTypeIteratorClass;
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
typedef struct _DbTypeIteratorClassRec 
{
  DbIRoutine clear;
  DbIRoutine first;
  DbIRoutine forth;
  DbIMoveRoutine move_to;
  DbIRoutine extend;
  DbIteratorIRoutine down;
  DbIteratorIRoutine up;
  DbIKeyRoutine get_key;
} DbTypeIteratorClassRec;

typedef struct _DbTypeIteratorRec 
{
  DbTypeIteratorClass classDef;

  DbTypeIterator parent;
  DbType type;
  void* address;
  IteratorMode mode;
  int finished;
} DbTypeIteratorRec;
/*---------------------------------------------------------------------*/

/*-- Static global variables. -----------------------------------------*/
static int DbActive    = 0;

static GCClass DbTypeGCClass      = NULL;
static GCClass DbFieldGCClass     = NULL;

static DbType  DbTCharObject      = NULL;
static DbType  DbTShortObject     = NULL;
static DbType  DbTIntObject       = NULL;
static DbType  DbTHexaObject      = NULL;
static DbType  DbTFloatObject     = NULL;
static DbType  DbTDoubleObject    = NULL;
static DbType  DbTStringObject    = NULL;

static int DbTypeIteratorCount  = 0;
static int DbObjectValueCount   = 0;
static int DbTypeCount          = 0;
static int DbStringOutputCount  = 0;
static int DbFieldCount         = 0;
static int DbFieldNameCount     = 0;
static int DbStructureNameCount = 0;

static int Debug = 0;
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
int DbDebug ();

static void DbFieldDelete (DbField This);
static void DbFieldClear (DbField This);

static void DbTypeDelete (DbType This);
static void DbTypeClear (DbType This);
static void DbTypeUninstall (DbType This);

static void DbTReferenceCleanup ();
static void DbTVectorCleanup ();
static void DbTMatrixCleanup ();
static void DbTStructureCleanup ();

static DbTypeIterator DbTypeIteratorNew (DbTypeIterator parent, 
                                         DbType type, 
                                         IteratorMode mode, 
                                         void* address);
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                                                                     */
/* TypeIterator management.                                            */
/*                                                                     */
/*---------------------------------------------------------------------*/

static DbTypeIteratorClassRec TypeIteratorClass = {
  (DbIRoutine) NULL,
  (DbIRoutine) NULL,
  (DbIRoutine) NULL,
  (DbIMoveRoutine) NULL,
  (DbIRoutine) NULL,
  (DbIteratorIRoutine) NULL,
  (DbIteratorIRoutine) NULL,
  (DbIKeyRoutine) NULL
};

/*- procedure ---------------------------------------------------------*/
static DbTypeIterator DbTypeIteratorNew (DbTypeIterator parent, 
                                         DbType type, 
                                         IteratorMode mode, 
                                         void* address)
/*---------------------------------------------------------------------*/
{
  DbTypeIterator This;

  if (type == NULL) return (NULL);

  This = (DbTypeIterator) malloc (sizeof (DbTypeIteratorRec));
  DbTypeIteratorCount++;

  This->classDef = &TypeIteratorClass;

  This->parent   = parent;
  This->type     = type;
  This->address  = address;
  This->mode     = mode;
  This->finished = 0;

  if (Debug)
    {
      fprintf (stderr, 
               "DbType> Create DefaultTypeIterator (type %s) #%d\n", 
               DbTypeName (type), DbTypeIteratorCount);
    }

  return (This);
}

/*- procedure ---------------------------------------------------------*/
void DbTypeIteratorDelete (DbTypeIterator This)
/*---------------------------------------------------------------------*/
{
  DbType type;
  DbIRoutine clear;

  if (This == NULL) return;

  type = This->type;
  clear = This->classDef->clear;
  if (clear != NULL)
    {
      clear (This);
    }

  This->parent = NULL;
  This->type = NULL;
  This->address = NULL;
  This->finished = 1;

  free (This);
  DbTypeIteratorCount--;

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Delete TypeIterator (type %s) #%d\n", 
               DbTypeName (type), DbTypeIteratorCount);
    }

}

/*- procedure ---------------------------------------------------------*/
int DbTypeIteratorFinished (DbTypeIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (1);

  return (This->finished);
}

/*- procedure ---------------------------------------------------------*/
DbTypeIterator DbTypeIteratorDown (DbTypeIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (NULL);

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Down TypeIterator (type %s)\n", 
               DbTypeName (This->type));
    }
      
  if (This->classDef->down != NULL)
    {
      return (This->classDef->down (This));
    }
  else
    {
      return (NULL);
    }
}

/*- procedure ---------------------------------------------------------*/
static void DbTypeIteratorSetParent (DbTypeIterator This, 
                                     DbTypeIterator parent)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  This->parent = parent;
}

/*- procedure ---------------------------------------------------------*/
DbTypeIterator DbTypeIteratorUp (DbTypeIterator This)
/*---------------------------------------------------------------------*/
{
  DbTypeIterator parent = NULL;

  if (This != NULL)
    {
      if (Debug)
        {
          fprintf (stderr,
                   "DbType> Up TypeIterator (type %s)\n", 
                   DbTypeName (This->type));
        }
      
      parent = This->parent;
      DbTypeIteratorDelete ((DbTypeIterator) This);
    }

  return (parent);
}

/*- procedure ---------------------------------------------------------*/
void DbTypeIteratorForth (DbTypeIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;
  if (This->finished) return;

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Forth TypeIterator (type %s ->", 
               DbTypeName (This->type));
    }

  if (This->classDef->forth != NULL)
    {
      This->classDef->forth (This);
    }
  else
    {
      This->finished = 1;
    }

  if (Debug)
    {
      fprintf (stderr,
               " %s)\n", 
               DbTypeName (This->type));
    }

}

/*- procedure ---------------------------------------------------------*/
void DbTypeIteratorMoveTo (DbTypeIterator This, char* key)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Move TypeIterator to %s (type %s ->", 
               key,
               DbTypeName (This->type));
    }

  if (This->classDef->move_to != NULL)
    {
      This->classDef->move_to (This, key);
    }

  if (Debug)
    {
      fprintf (stderr,
               " %s)\n", 
               DbTypeName (This->type));
    }

}

/*- procedure ---------------------------------------------------------*/
void DbTypeIteratorExtend (DbTypeIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;
  if (This->finished) return;

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Extend TypeIterator (type %s)\n", 
               DbTypeName (This->type));
    }

  if (This->classDef->extend != NULL)
    {
      This->classDef->extend (This);
    }
}

/*- procedure ---------------------------------------------------------*/
void* DbTypeIteratorGetItemAddress (DbTypeIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (NULL);
  if (This->finished) return (NULL);

  return (This->address);
}

/*- procedure ---------------------------------------------------------*/
DbType DbTypeIteratorGetItemType (DbTypeIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (NULL);
  if (This->finished) return (NULL);

  return (This->type);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIteratorStoreChar (DbTypeIterator This, char value)
/*---------------------------------------------------------------------*/
{
  DbType type;

  if (This == NULL) return (0);
  if (This->finished) return (0);
  if (This->mode != write_only) return (0);

  if (This->address == NULL)
    {
      DbTypeIteratorExtend (This);
    }

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Store char %c TypeIterator (type %s)\n", value,
               DbTypeName (This->type));
    }

  type = This->type;

  if (type == DbTCharObject)
    {
      *((char*) This->address) = (char) value;
    }
  else if (type == DbTShortObject)
    {
      *((short*) This->address) = (short) value;
    }
  else if (type == DbTIntObject)
    {
      *((int*) This->address) = (int) value;
    }
  else if (type == DbTHexaObject)
    {
      *((hexa*) This->address) = (hexa) value;
    }
  else
    {
      return (0);
    }

  return (1);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIteratorStoreShort (DbTypeIterator This, short value)
/*---------------------------------------------------------------------*/
{
  DbType type;

  if (This == NULL) return (0);
  if (This->finished) return (0);
  if (This->mode != write_only) return (0);

  if (This->address == NULL)
    {
      DbTypeIteratorExtend (This);
    }

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Store short %hd TypeIterator (type %s)\n", value,
               DbTypeName (This->type));
    }

  type = This->type;

  if (type == DbTCharObject)
    {
      *((char*) This->address) = (char) value;
    }
  else if (type == DbTShortObject)
    {
      *((short*) This->address) = (short) value;
    }
  else if (type == DbTIntObject)
    {
      *((int*) This->address) = (int) value;
    }
  else if (type == DbTHexaObject)
    {
      *((hexa*) This->address) = (hexa) value;
    }
  else if (type == DbTFloatObject)
    {
      *((float*) This->address) = (float) value;
    }
  else if (type == DbTDoubleObject)
    {
      *((double*) This->address) = (double) value;
    }
  else
    {
      return (0);
    }

  return (1);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIteratorStoreInt (DbTypeIterator This, int value)
/*---------------------------------------------------------------------*/
{
  DbType type;

  if (This == NULL) return (0);
  if (This->finished) return (0);
  if (This->mode != write_only) return (0);

  if (This->address == NULL)
    {
      DbTypeIteratorExtend (This);
    }

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Store int %d TypeIterator (type %s)\n", value,
               DbTypeName (This->type));
    }

  type = This->type;

  if (type == DbTCharObject)
    {
      *((char*) This->address) = (char) value;
    }
  else if (type == DbTShortObject)
    {
      *((short*) This->address) = (short) value;
    }
  else if (type == DbTIntObject)
    {
      *((int*) This->address) = (int) value;
    }
  else if (type == DbTHexaObject)
    {
      *((hexa*) This->address) = (hexa) value;
    }
  else if (type == DbTFloatObject)
    {
      *((float*) This->address) = (float) value;
    }
  else if (type == DbTDoubleObject)
    {
      *((double*) This->address) = (double) value;
    }
  else
    {
      return (0);
    }

  return (1);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIteratorStoreFloat (DbTypeIterator This, float value)
/*---------------------------------------------------------------------*/
{
  DbType type;

  if (This == NULL) return (0);
  if (This->finished) return (0);
  if (This->mode != write_only) return (0);

  if (This->address == NULL)
    {
      DbTypeIteratorExtend (This);
    }

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Store float %f TypeIterator (type %s)\n", value,
               DbTypeName (This->type));
    }

  type = This->type;

  if (type == DbTCharObject)
    {
      *((char*) This->address) = (char) value;
    }
  else if (type == DbTShortObject)
    {
      *((short*) This->address) = (short) value;
    }
  else if (type == DbTIntObject)
    {
      *((int*) This->address) = (int) value;
    }
  else if (type == DbTHexaObject)
    {
      *((hexa*) This->address) = (hexa) value;
    }
  else if (type == DbTFloatObject)
    {
      *((float*) This->address) = (float) value;
    }
  else if (type == DbTDoubleObject)
    {
      *((double*) This->address) = (double) value;
    }
  else
    {
      return (0);
    }

  return (1);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIteratorStoreDouble (DbTypeIterator This, double value)
/*---------------------------------------------------------------------*/
{
  DbType type;

  if (This == NULL) return (0);
  if (This->finished) return (0);
  if (This->mode != write_only) return (0);

  if (This->address == NULL)
    {
      DbTypeIteratorExtend (This);
    }

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Store double %lg TypeIterator (type %s)\n", value,
               DbTypeName (This->type));
    }

  type = This->type;

  if (type == DbTCharObject)
    {
      *((char*) This->address) = (char) value;
    }
  else if (type == DbTShortObject)
    {
      *((short*) This->address) = (short) value;
    }
  else if (type == DbTIntObject)
    {
      *((int*) This->address) = (int) value;
    }
  else if (type == DbTHexaObject)
    {
      *((hexa*) This->address) = (hexa) value;
    }
  else if (type == DbTFloatObject)
    {
      *((float*) This->address) = (float) value;
    }
  else if (type == DbTDoubleObject)
    {
      *((double*) This->address) = (double) value;
    }
  else
    {
      return (0);
    }

  return (1);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIteratorStoreText (DbTypeIterator This, char* value)
/*---------------------------------------------------------------------*/
{
  DbType type;
  DbString* string;

  if (This == NULL) return (0);
  if (This->finished) return (0);
  if (This->mode != write_only) return (0);

  if (This->address == NULL)
    {
      DbTypeIteratorExtend (This);
    }

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Store text %s TypeIterator (type %s)\n", value,
               DbTypeName (This->type));
    }

  type = This->type;

  if (type != DbTStringObject) return (0);

  string = (DbString*) This->address;

  DbStringSetContents (string, value);

  return (1);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIteratorStoreName (DbTypeIterator This, char* value)
/*---------------------------------------------------------------------*/
{
  DbType type;
  DbReference* reference;

  if (This == NULL) return (0);
  if (This->finished) return (0);
  if (This->mode != write_only) return (0);

  if (This->address == NULL)
    {
      DbTypeIteratorExtend (This);
    }

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Store name %s TypeIterator (type %s)\n", value,
               DbTypeName (This->type));
    }

  type = This->type;

  if (!DbTypeIsReference (type)) return (0);

  reference = (DbReference*) This->address;

  DbReferenceSetName (reference, value);

  return (1);
}

/*- procedure ---------------------------------------------------------*/
char* DbTypeIteratorGetKey (DbTypeIterator This)
/*---------------------------------------------------------------------*/
{
  char* result = NULL;

  if (This == NULL) return (NULL);

  if (Debug)
    {
      fprintf (stderr,
               "DbType> GetKey on TypeIterator (type %s ->", 
               DbTypeName (This->type));
    }

  if (This->classDef->get_key != NULL)
    {
      result = This->classDef->get_key (This);
    }

  if (Debug)
    {
      fprintf (stderr,
               " %s) -> %s\n", 
               DbTypeName (This->type), result);
    }

  return (result);
}

/*---------------------------------------------------------------------*/
/*                                                                     */
/*  Type  management.                                                  */
/*                                                                     */
/*---------------------------------------------------------------------*/


/*- procedure ---------------------------------------------------------*/
void DbTypeStartup ()
/*---------------------------------------------------------------------*/
{
  if (!DbActive)
    {
      DbActive = 1;
      
      GCReferenceAssign (DbTypeGCClass, 
                         GCClassNew ("DbType", (GCDeleteMethod) DbTypeDelete));
      GCReferenceAssign (DbFieldGCClass, 
                         GCClassNew ("DbField", 
                                     (GCDeleteMethod) DbFieldDelete));
    }
}

/*- procedure ---------------------------------------------------------*/
void DbTypeCleanup ()
/*---------------------------------------------------------------------*/
{
  if (DbActive)
    {
      DbActive = 0;

      DbTReferenceCleanup ();
      DbTVectorCleanup ();
      DbTMatrixCleanup ();
      DbTStructureCleanup ();

      GCReferenceClear (DbTCharObject);
      GCReferenceClear (DbTShortObject);
      GCReferenceClear (DbTIntObject);
      GCReferenceClear (DbTHexaObject);
      GCReferenceClear (DbTFloatObject);
      GCReferenceClear (DbTDoubleObject);
      GCReferenceClear (DbTStringObject);

      GCReferenceClear (DbTypeGCClass);
      GCReferenceClear (DbFieldGCClass);
    }
}

/*- procedure ---------------------------------------------------------*/
static void DbTypeDelete (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  if (DbDebug ()) printf ("DbTypeDelete> type = %lx\n", This);

  DbTypeClear (This);

  free (This);
  DbTypeCount--;
}

/*- procedure ---------------------------------------------------------*/
static void DbTypeClear (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  if (This->classDef->Clear) This->classDef->Clear (This);
}

/*- procedure ---------------------------------------------------------*/
char* DbTypeName (DbType This)
/*---------------------------------------------------------------------*/
{
  static char nullName[] = "";
  
  if (This == NULL) return (nullName);
  if (This->classDef->Name == NULL) return (nullName);
  return ((char*) This->classDef->Name (This));
}

/*- procedure ---------------------------------------------------------*/
char* DbTypeCName (DbType This)
/*---------------------------------------------------------------------*/
{
  static char nullName[] = "";
  
  if (This == NULL) return (nullName);
  if (This->classDef->CName == NULL)
    {
      return (DbTypeName (This));
    }
  else
    {
      return ((char*) This->classDef->CName (This));
    }
}

/*- procedure ---------------------------------------------------------*/
char* DbTypeCppName (DbType This)
/*---------------------------------------------------------------------*/
{
  static char nullName[] = "";
  
  if (This == NULL) return (nullName);
  if (This->classDef->CppName == NULL)
    {
      return (DbTypeCName (This));
    }
  else
    {
      return ((char*) This->classDef->CppName (This));
    }
}

/*- procedure ---------------------------------------------------------*/
char* DbTypeSubTypeName (DbType This)
/*---------------------------------------------------------------------*/
{
  static char nullName[] = "";
  
  if (This == NULL) return (nullName);
  if (This->classDef->SubTypeName == NULL)
    {
      return (nullName);
    }
  else
    {
      return ((char*) This->classDef->SubTypeName (This));
    }
}

/*- procedure ---------------------------------------------------------*/
int DbTypeSize (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (0);
  return (This->classDef->Size (This));
}

/*- procedure ---------------------------------------------------------*/
static int DbTypeAlignment (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (0);
  return (This->classDef->Alignment (This));
}

/*- procedure ---------------------------------------------------------*/
static int DbTypeAlign (DbType This, int offset)
/*---------------------------------------------------------------------*/
{
  int alignment;

  if (This == NULL) return (0);
  alignment = DbTypeAlignment(This);

  if (alignment <= 1) return (0);

  if (!(offset % alignment)) return (0);
  return ((offset & ~(alignment - 1)) + alignment - offset);
}

/*- procedure ---------------------------------------------------------*/
char* DbTypeOutput (DbType This, void* address)
/*---------------------------------------------------------------------*/
{
  static char nullText[] = "";
  
  if (This == NULL) return (nullText);
  if (address == NULL) return (nullText);

  if (This->classDef->Output == NULL) return (nullText);
  
  return ((char*) This->classDef->Output (This, address));
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIsSame (DbType This, DbType other)
/*---------------------------------------------------------------------*/
{
  if (This == other) return (1);
  if (This == NULL) return (0);
  if (other == NULL) return (0);
  if (This->classDef != other->classDef) return (0);

    /*
      At this point, This and other are of the same class.
      */

  if (This->classDef->IsSame != NULL)
    {
      return (This->classDef->IsSame (This, other));
    }
  else
    {
      return (1);
    }
}

/*- procedure ---------------------------------------------------------*/
DbType DbTypeGetUsedStructure (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (NULL);

  if (This->classDef->UsedStructure != NULL)
    {
      return (This->classDef->UsedStructure (This));
    }
  else
    {
      return (NULL);
    }
}

/*- procedure ---------------------------------------------------------*/
DbTypeIterator DbTypeGiveInputIterator (DbType This, void* address)
/*---------------------------------------------------------------------*/
{
  DbIteratorRoutine f;

  if (This == NULL) return (NULL);
  if (address == NULL) return (NULL);

  f = This->classDef->GiveIterator;

  if (f == NULL)
    {
      return (DbTypeIteratorNew (NULL, This, read_only, address));
    }
  else
    {
      return (f (This, read_only, address));
    }
}

/*- procedure ---------------------------------------------------------*/
DbTypeIterator DbTypeGiveOutputIterator (DbType This, void* address)
/*---------------------------------------------------------------------*/
{
  DbIteratorRoutine f;

  if (This == NULL) return (NULL);
  if (address == NULL) return (NULL);

  f = This->classDef->GiveIterator;

  if (f == NULL)
    {
      return (DbTypeIteratorNew (NULL, This, write_only, address));
    }
  else
    {
      return (f (This, write_only, address));
    }
}

/*- procedure ---------------------------------------------------------*/
DbTypeIterator DbTypeGiveVisitIterator (DbType This)
/*---------------------------------------------------------------------*/
{
  DbIteratorRoutine f;

  if (This == NULL) return (NULL);

  f = This->classDef->GiveIterator;

  if (f == NULL)
    {
      return (DbTypeIteratorNew (NULL, This, visit_only, NULL));
    }
  else
    {
      return (f (This, visit_only, NULL));
    }
}

/*- procedure ---------------------------------------------------------*/
void* DbTypeInstanciate (DbType This)
/*---------------------------------------------------------------------*/
{
  void* address;

  if (This == NULL) return (NULL);

  if (This->classDef->Instanciate == NULL)
    {
      int size;

      size = DbTypeSize (This);

      address = (void*) malloc (size);
      DbObjectValueCount++;

      DbTypeInit (This, address);
    }
  else
    {
      address = This->classDef->Instanciate (This);
    }

  return (address);
}
                                                                        
/*- procedure ---------------------------------------------------------*/
void DbTypeInit (DbType This, void* address)
/*---------------------------------------------------------------------*/
{
  int size;

  if (This == NULL) return;
  if (address == NULL) return;

  if (This->classDef->Init == NULL)
    {
      size = DbTypeSize (This);

      memset (address, 0, size);
    }
  else
    {
      This->classDef->Init (This, address);
    }
}
                                                                        
/*- procedure ---------------------------------------------------------*/
void DbTypeErase (DbType This, void* address)
/*---------------------------------------------------------------------*/
{
  int size;

  if (This == NULL) return;
  if (address == NULL) return;

  if (This->classDef->Erase == NULL)
    {
      size = DbTypeSize (This);

      memset (address, 0, size);
    }
  else
    {
      This->classDef->Erase (This, address);
    }
}
                                                                        
/*- procedure ---------------------------------------------------------*/
static void DbTypeCheck ()
/*---------------------------------------------------------------------*/
{
  printf ("itera=%d obval=%d types=%d strou=%d field=%d fldna=%d "
          "strnm=%d\n",
          DbTypeIteratorCount,
          DbObjectValueCount,
          DbTypeCount,
          DbStringOutputCount,
          DbFieldCount,
          DbFieldNameCount,
          DbStructureNameCount);
}

/*---------------------------------------------------------------------*/
/*                                                                     */
/* Char   management.                                                  */
/*                                                                     */
/*---------------------------------------------------------------------*/

static char* DbTCharName (DbTChar type);
static int DbTCharSize (DbTChar type);
static int DbTCharAlignment (DbTChar type);
static char* DbTCharOutput (DbTChar type, void* address);

static DbTypeClassRec DbTCharClass = {
  "DbTChar",

  (DbRoutine) NULL,
  (DbStringRoutine) DbTCharName,
  (DbStringRoutine) NULL,
  (DbStringRoutine) NULL,
  (DbStringRoutine) NULL,
  (DbIntRoutine) DbTCharSize,
  (DbIntRoutine) DbTCharAlignment,
  (DbPtrRoutine) NULL,
  (DbInitRoutine) NULL,
  (DbInitRoutine) NULL,
  (DbIteratorRoutine) NULL,
  (DbOutputRoutine) DbTCharOutput,
  (DbCompareRoutine) NULL,
  (DbUsedStructureRoutine) NULL
};

/*- procedure ---------------------------------------------------------*/
DbType DbTCharNew ()
/*---------------------------------------------------------------------*/
{
  DbType type = DbTCharObject;

  if (type == NULL)
    {
      DbTypeStartup ();

      type = (DbType) malloc (sizeof(DbTypeRec));
      DbTypeCount++;

      type->classDef = &DbTCharClass;

      GCObjectNew (type, DbTypeGCClass);
      GCReferenceAssign (DbTCharObject, type);
    }

  return ((DbType) (type));
}

/*- procedure ---------------------------------------------------------*/
static char* DbTCharName (DbTChar type)
/*---------------------------------------------------------------------*/
{
  return ("char");
}

/*- procedure ---------------------------------------------------------*/
static int DbTCharSize (DbTChar type)
/*---------------------------------------------------------------------*/
{
  return (sizeof (char));
}

/*- procedure ---------------------------------------------------------*/
static int DbTCharAlignment (DbTChar type)
/*---------------------------------------------------------------------*/
{
  DbBuildAlignStruct (char);

  return (DbComputeAlign);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTCharOutput (DbTChar type, void* address)
/*---------------------------------------------------------------------*/
{
  static char text[10];
  char value = *((char*) address);

  if (iscntrl (value))
    {
      sprintf (text, "'\\%3.3o'", value);
    }
  else
    {
      sprintf (text, "'%c'", value);
    }

  return ((char*) text);
}

/*---------------------------------------------------------------------*/
/*                                                                     */
/* Short   management.                                                 */
/*                                                                     */
/*---------------------------------------------------------------------*/

static char* DbTShortName (DbTShort type);
static int DbTShortSize (DbTShort type);
static int DbTShortAlignment (DbTShort type);
static char* DbTShortOutput (DbTShort type, void* address);

static DbTypeClassRec DbTShortClass = {
  "DbTShort",

  (DbRoutine) NULL,
  (DbStringRoutine) DbTShortName,
  (DbStringRoutine) NULL,
  (DbStringRoutine) NULL,
  (DbStringRoutine) NULL,
  (DbIntRoutine) DbTShortSize,
  (DbIntRoutine) DbTShortAlignment,
  (DbPtrRoutine) NULL,
  (DbInitRoutine) NULL,
  (DbInitRoutine) NULL,
  (DbIteratorRoutine) NULL,
  (DbOutputRoutine) DbTShortOutput,
  (DbCompareRoutine) NULL,
  (DbUsedStructureRoutine) NULL
};

/*- procedure ---------------------------------------------------------*/
DbType DbTShortNew ()
/*---------------------------------------------------------------------*/
{
  DbType type = DbTShortObject;

  if (type == NULL)
    {
      DbTypeStartup ();

      type = (DbType) malloc (sizeof(DbTypeRec));
      DbTypeCount++;

      type->classDef = &DbTShortClass;

      GCObjectNew (type, DbTypeGCClass);
      GCReferenceAssign (DbTShortObject, type);
    }

  return ((DbType) (type));
}

/*- procedure ---------------------------------------------------------*/
static char* DbTShortName (DbTShort type)
/*---------------------------------------------------------------------*/
{
  return ("short");
}

/*- procedure ---------------------------------------------------------*/
static int DbTShortSize (DbTShort type)
/*---------------------------------------------------------------------*/
{
  return (sizeof (short));
}

/*- procedure ---------------------------------------------------------*/
static int DbTShortAlignment (DbTShort type)
/*---------------------------------------------------------------------*/
{
  DbBuildAlignStruct (short);

  return (DbComputeAlign);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTShortOutput (DbTShort type, void* address)
/*---------------------------------------------------------------------*/
{
  static char text[20];
  short value = *((short*) address);

  sprintf (text, "%hd", value);

  return ((char*) text);
}

/*---------------------------------------------------------------------*/
/*                                                                     */
/* Int   management.                                                   */
/*                                                                     */
/*---------------------------------------------------------------------*/

static char* DbTIntName (DbTInt type);
static int DbTIntSize (DbTInt type);
static int DbTIntAlignment (DbTInt type);
static char* DbTIntOutput (DbTInt type, void* address);

static DbTypeClassRec DbTIntClass = {
  "DbTInt",

  (DbRoutine) NULL,
  (DbStringRoutine) DbTIntName,
  (DbStringRoutine) NULL,
  (DbStringRoutine) NULL,
  (DbStringRoutine) NULL,
  (DbIntRoutine) DbTIntSize,
  (DbIntRoutine) DbTIntAlignment,
  (DbPtrRoutine) NULL,
  (DbInitRoutine) NULL,
  (DbInitRoutine) NULL,
  (DbIteratorRoutine) NULL,
  (DbOutputRoutine) DbTIntOutput,
  (DbCompareRoutine) NULL,
  (DbUsedStructureRoutine) NULL
};

/*- procedure ---------------------------------------------------------*/
DbType DbTIntNew ()
/*---------------------------------------------------------------------*/
{
  DbType type = DbTIntObject;

  if (type == NULL)
    {
      DbTypeStartup ();

      type = (DbType) malloc (sizeof(DbTypeRec));
      DbTypeCount++;

      type->classDef = &DbTIntClass;

      GCObjectNew (type, DbTypeGCClass);
      GCReferenceAssign (DbTIntObject, type);
    }

  return ((DbType) (type));
}

/*- procedure ---------------------------------------------------------*/
static char* DbTIntName (DbTInt type)
/*---------------------------------------------------------------------*/
{
  return ("int");
}

/*- procedure ---------------------------------------------------------*/
static int DbTIntSize (DbTInt type)
/*---------------------------------------------------------------------*/
{
  return (sizeof (int));
}

/*- procedure ---------------------------------------------------------*/
static int DbTIntAlignment (DbTInt type)
/*---------------------------------------------------------------------*/
{
  DbBuildAlignStruct (int);

  return (DbComputeAlign);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTIntOutput (DbTInt type, void* address)
/*---------------------------------------------------------------------*/
{
  static char text[20];
  int value = *((int*) address);

  sprintf (text, "%d", value);

  return ((char*) text);
}

/*---------------------------------------------------------------------*/
/*                                                                     */
/* Hexa   management.                                                  */
/*                                                                     */
/*---------------------------------------------------------------------*/

static char* DbTHexaName (DbTHexa type);
static char* DbTHexaCName (DbTHexa This);
static int DbTHexaSize (DbTHexa type);
static int DbTHexaAlignment (DbTHexa type);
static char* DbTHexaOutput (DbTHexa type, void* address);

static DbTypeClassRec DbTHexaClass = {
  "DbTHexa",

  (DbRoutine) NULL,
  (DbStringRoutine) DbTHexaName,
  (DbStringRoutine) DbTHexaCName,
  (DbStringRoutine) NULL,
  (DbStringRoutine) NULL,
  (DbIntRoutine) DbTHexaSize,
  (DbIntRoutine) DbTHexaAlignment,
  (DbPtrRoutine) NULL,
  (DbInitRoutine) NULL,
  (DbInitRoutine) NULL,
  (DbIteratorRoutine) NULL,
  (DbOutputRoutine) DbTHexaOutput,
  (DbCompareRoutine) NULL,
  (DbUsedStructureRoutine) NULL
};

/*- procedure ---------------------------------------------------------*/
DbType DbTHexaNew ()
/*---------------------------------------------------------------------*/
{
  DbType type = DbTHexaObject;

  if (type == NULL)
    {
      DbTypeStartup ();

      type = (DbType) malloc (sizeof(DbTypeRec));
      DbTypeCount++;

      type->classDef = &DbTHexaClass;

      GCObjectNew (type, DbTypeGCClass);
      GCReferenceAssign (DbTHexaObject, type);
    }

  return ((DbType) (type));
}

/*- procedure ---------------------------------------------------------*/
static char* DbTHexaCName (DbTHexa type)
/*---------------------------------------------------------------------*/
{
  return ("unsigned int");
}

/*- procedure ---------------------------------------------------------*/
static char* DbTHexaName (DbTHexa type)
/*---------------------------------------------------------------------*/
{
  return ("hexa");
}

/*- procedure ---------------------------------------------------------*/
static int DbTHexaSize (DbTHexa type)
/*---------------------------------------------------------------------*/
{
  return (sizeof (hexa));
}

/*- procedure ---------------------------------------------------------*/
static int DbTHexaAlignment (DbTHexa type)
/*---------------------------------------------------------------------*/
{
  DbBuildAlignStruct (hexa);

  return (DbComputeAlign);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTHexaOutput (DbTHexa type, void* address)
/*---------------------------------------------------------------------*/
{
  static char text[20];
  hexa value = *((hexa*) address);

  sprintf (text, "0x%x", value);

  return ((char*) text);
}

/*---------------------------------------------------------------------*/
/*                                                                     */
/* Float   management.                                                 */
/*                                                                     */
/*---------------------------------------------------------------------*/

static char* DbTFloatName (DbTFloat type);
static int DbTFloatSize (DbTFloat type);
static int DbTFloatAlignment (DbTFloat type);
static char* DbTFloatOutput (DbTFloat type, void* address);

static DbTypeClassRec DbTFloatClass = {
  "DbTFloat",

  (DbRoutine) NULL,
  (DbStringRoutine) DbTFloatName,
  (DbStringRoutine) NULL,
  (DbStringRoutine) NULL,
  (DbStringRoutine) NULL,
  (DbIntRoutine) DbTFloatSize,
  (DbIntRoutine) DbTFloatAlignment,
  (DbPtrRoutine) NULL,
  (DbInitRoutine) NULL,
  (DbInitRoutine) NULL,
  (DbIteratorRoutine) NULL,
  (DbOutputRoutine) DbTFloatOutput,
  (DbCompareRoutine) NULL,
  (DbUsedStructureRoutine) NULL
};

/*- procedure ---------------------------------------------------------*/
DbType DbTFloatNew ()
/*---------------------------------------------------------------------*/
{
  DbType type = DbTFloatObject;

  if (type == NULL)
    {
      DbTypeStartup ();

      type = (DbType) malloc (sizeof(DbTypeRec));
      DbTypeCount++;

      type->classDef = &DbTFloatClass;

      GCObjectNew (type, DbTypeGCClass);
      GCReferenceAssign (DbTFloatObject, type);
    }

  return ((DbType) (type));
}

/*- procedure ---------------------------------------------------------*/
static char* DbTFloatName (DbTFloat type)
/*---------------------------------------------------------------------*/
{
  return ("float");
}

/*- procedure ---------------------------------------------------------*/
static int DbTFloatSize (DbTFloat type)
/*---------------------------------------------------------------------*/
{
  return (sizeof (float));
}

/*- procedure ---------------------------------------------------------*/
static int DbTFloatAlignment (DbTFloat type)
/*---------------------------------------------------------------------*/
{
  DbBuildAlignStruct (float);

  return (DbComputeAlign);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTFloatOutput (DbTFloat type, void* address)
/*---------------------------------------------------------------------*/
{
  static char text[40];
  float value = *((float*) address);

  /*  sprintf (text, "%.16f", value);*/
  sprintf (text, "%e", value);

  return ((char*) text);
}

/*---------------------------------------------------------------------*/
/*                                                                     */
/* Double   management.                                                */
/*                                                                     */
/*---------------------------------------------------------------------*/

static char* DbTDoubleName (DbTDouble type);
static int DbTDoubleSize (DbTDouble type);
static int DbTDoubleAlignment (DbTDouble type);
static char* DbTDoubleOutput (DbTDouble type, void* address);

static DbTypeClassRec DbTDoubleClass = {
  "DbTDouble",

  (DbRoutine) NULL,
  (DbStringRoutine) DbTDoubleName,
  (DbStringRoutine) NULL,
  (DbStringRoutine) NULL,
  (DbStringRoutine) NULL,
  (DbIntRoutine) DbTDoubleSize,
  (DbIntRoutine) DbTDoubleAlignment,
  (DbPtrRoutine) NULL,
  (DbInitRoutine) NULL,
  (DbInitRoutine) NULL,
  (DbIteratorRoutine) NULL,
  (DbOutputRoutine) DbTDoubleOutput,
  (DbCompareRoutine) NULL,
  (DbUsedStructureRoutine) NULL
};

/*- procedure ---------------------------------------------------------*/
DbType DbTDoubleNew ()
/*---------------------------------------------------------------------*/
{
  DbType type = DbTDoubleObject;

  if (type == NULL)
    {
      DbTypeStartup ();

      type = (DbType) malloc (sizeof(DbTypeRec));
      DbTypeCount++;

      type->classDef = &DbTDoubleClass;

      GCObjectNew (type, DbTypeGCClass);
      GCReferenceAssign (DbTDoubleObject, type);
    }

  return ((DbType) (type));
}

/*- procedure ---------------------------------------------------------*/
static char* DbTDoubleName (DbTDouble type)
/*---------------------------------------------------------------------*/
{
  return ("double");
}

/*- procedure ---------------------------------------------------------*/
static int DbTDoubleSize (DbTDouble type)
/*---------------------------------------------------------------------*/
{
  return (sizeof (double));
}

/*- procedure ---------------------------------------------------------*/
static int DbTDoubleAlignment (DbTDouble type)
/*---------------------------------------------------------------------*/
{
  DbBuildAlignStruct (double);

  return (DbComputeAlign);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTDoubleOutput (DbTDouble type, void* address)
/*---------------------------------------------------------------------*/
{
  static char text[40];
  double value = *((double*) address);

  /*  sprintf (text, "%.16lg", value);*/
  sprintf (text, "%e", value);

  return ((char*) text);
}

/*---------------------------------------------------------------------*/
/*                                                                     */
/* String  management.                                                 */
/*                                                                     */
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
typedef struct _DbTStringRec {
  DbTypeRec base;
} DbTStringRec;

static char* DbTStringName (DbTString This);
static char* DbTStringCName (DbTString This);
static int DbTStringSize (DbTString This);
static int DbTStringAlignment (DbTString This);
static char* DbTStringOutput (DbTString This, void* address);
static void* DbTStringInstanciate (DbTString This);
static void DbTStringInit (DbTString This, void* address);
static void DbTStringErase (DbTString This, void* address);

static DbTypeClassRec DbTStringClass = {
  "DbTString",

  (DbRoutine) NULL,
  (DbStringRoutine) DbTStringName,
  (DbStringRoutine) DbTStringCName,
  (DbStringRoutine) NULL,
  (DbStringRoutine) NULL,
  (DbIntRoutine) DbTStringSize,
  (DbIntRoutine) DbTStringAlignment,
  (DbPtrRoutine) DbTStringInstanciate,
  (DbInitRoutine) DbTStringInit,
  (DbInitRoutine) DbTStringErase,
  (DbIteratorRoutine) NULL,
  (DbOutputRoutine) DbTStringOutput,
  (DbCompareRoutine) NULL,
  (DbUsedStructureRoutine) NULL
};
/*---------------------------------------------------------------------*/

/*- procedure ---------------------------------------------------------*/
DbType DbTStringNew ()
/*---------------------------------------------------------------------*/
{
  DbType type = DbTStringObject;

  if (type == NULL)
    {
      DbTypeStartup ();

      type = (DbType) malloc (sizeof(DbTypeRec));
      DbTypeCount++;

      type->classDef = &DbTStringClass;

      GCObjectNew (type, DbTypeGCClass);
      GCReferenceAssign (DbTStringObject, type);
    }

  return ((DbType) (type));
}

/*- procedure ---------------------------------------------------------*/
static char* DbTStringName (DbTString This)
/*---------------------------------------------------------------------*/
{
  return ("string");
}

/*- procedure ---------------------------------------------------------*/
static char* DbTStringCName (DbTString This)
/*---------------------------------------------------------------------*/
{
  return ("DbString");
}

/*- procedure ---------------------------------------------------------*/
static int DbTStringSize (DbTString This)
/*---------------------------------------------------------------------*/
{
  return (sizeof (DbString));
}

/*- procedure ---------------------------------------------------------*/
static int DbTStringAlignment (DbTString This)
/*---------------------------------------------------------------------*/
{
  DbBuildAlignStruct (DbString);

  return (DbComputeAlign);
}

/*- procedure ---------------------------------------------------------*/
static void* DbTStringInstanciate (DbTString This)
/*---------------------------------------------------------------------*/
{
  return (DbStringNew ());
}

/*- procedure ---------------------------------------------------------*/
static void DbTStringInit (DbTString This, void* address)
/*---------------------------------------------------------------------*/
{
  DbString* string = (DbString*) address;

  DbStringInit (string);
}

/*- procedure ---------------------------------------------------------*/
static void DbTStringErase (DbTString This, void* address)
/*---------------------------------------------------------------------*/
{
  DbString* string = (DbString*) address;

  DbStringErase (string);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTStringOutput (DbTString type, void* address)
/*---------------------------------------------------------------------*/
{
  static char* empty = "\"\"";
  static char* text = NULL;
  DbString* string = (DbString*) address;
  char* value;

  if (string == NULL) return (empty);
  
  value = string->contents;
  if (value == NULL) return (empty);

  if (text == NULL)
    {
      text = (char*) malloc (strlen (value) + 3);
      DbStringOutputCount++;
    }
  else
    {
      text = (char*) realloc (text, strlen (value) + 3);
    }

  sprintf (text, "\"%s\"", value);

  return ((char*) text);
}

/*---------------------------------------------------------------------*/
/*                                                                     */
/* Reference  management.                                              */
/*                                                                     */
/*---------------------------------------------------------------------*/

typedef struct _DbTReferenceRec {
  DbTypeRec base;
  DbType    element;
} DbTReferenceRec;

static void DbTReferenceClear (DbTReference This);
static char* DbTReferenceName (DbTReference This);
static char* DbTReferenceCName (DbTReference This);
static char* DbTReferenceCppName (DbTReference This);
static char* DbTReferenceSubTypeName (DbTReference This);
static int DbTReferenceSize (DbTReference This);
static int DbTReferenceAlignment (DbTReference This);
static void* DbTReferenceInstanciate (DbTReference This);
static void DbTReferenceInit (DbTReference This, void* address);
static void DbTReferenceErase (DbTReference This, void* address);
static char* DbTReferenceOutput (DbTReference type, void* address);
static int DbTReferenceIsSame (DbTReference This, DbTReference other);
static DbType DbTReferenceUsedStructure (DbTReference This);

static DbTypeClassRec DbTReferenceClass = {
  "DbTReference",

  (DbRoutine) DbTReferenceClear,
  (DbStringRoutine) DbTReferenceName,
  (DbStringRoutine) DbTReferenceCName,
  (DbStringRoutine) DbTReferenceCppName,
  (DbStringRoutine) DbTReferenceSubTypeName,
  (DbIntRoutine) DbTReferenceSize,
  (DbIntRoutine) DbTReferenceAlignment,
  (DbPtrRoutine) DbTReferenceInstanciate,
  (DbInitRoutine) DbTReferenceInit,
  (DbInitRoutine) DbTReferenceErase,
  (DbIteratorRoutine) NULL,
  (DbOutputRoutine) DbTReferenceOutput,
  (DbCompareRoutine) DbTReferenceIsSame,
  (DbUsedStructureRoutine) DbTReferenceUsedStructure
};

/*---------------------------------------------------------------------*/
static CLList References = NULL;
/*---------------------------------------------------------------------*/

/*- procedure ---------------------------------------------------------*/
static void DbTReferenceCleanup ()
/*---------------------------------------------------------------------*/
{
  CLListErase (References);
  GCReferenceClear (References);

  GCCleanup ();
}

/*- procedure ---------------------------------------------------------*/
static DbType DbTReferenceFind (DbType element)
/*---------------------------------------------------------------------*/
{
  CIterator it;
  DbType result = NULL;

  if (References == NULL) return (NULL);

  it = CLListGiveIterator (References);
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbTReference r;

      r = (DbTReference) CIteratorItem (it);
      if (r != NULL)
        {
          if (r->element == element)
            {
              result = (DbType) r;
              break;
            }
        }
    }
  CIteratorDelete (it);
      
  return (result);
}

/*- procedure ---------------------------------------------------------*/
DbType DbTReferenceNew (DbType element)
/*---------------------------------------------------------------------*/
{
  DbTReference This;

  This = (DbTReference) DbTReferenceFind (element);
  if (This != NULL) return ((DbType) This);

  DbTypeStartup ();

  This = (DbTReference) malloc (sizeof (DbTReferenceRec));
  DbTypeCount++;

  This->base.classDef = &DbTReferenceClass;
    
  This->element  = NULL;

  GCObjectNew (This, DbTypeGCClass);

  GCReferenceAssignIn (This->element, This, element);

  if (References == NULL)
    {
      GCReferenceAssign (References, CLListNew ());
    }

  CLListAdd (References, This);

  return ((DbType) This);
}

/*- procedure ---------------------------------------------------------*/
static void DbTReferenceUninstall (DbTReference This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  if (DbTReferenceFind (This->element) == NULL) return;
  CLListRemove (References, This);

  DbTypeUninstall ((DbType) This);
}

/*- procedure ---------------------------------------------------------*/
static void DbTReferenceClear (DbTReference This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  GCReferenceClear (This->element);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTReferenceName (DbTReference This)
/*---------------------------------------------------------------------*/
{
  static char text[256];
  char temp[240];
    
  char* name;
    
  strcpy (temp, "reference");

  if (This->element != NULL)
    {
      strcat (temp, "<");
      name = DbTypeName (This->element);
      if (strlen (name) > sizeof(temp))
        {
            /* strcat (temp, "Strange type name"); */
        }
      else
        {
          strcat (temp, name);
        }
      strcat (temp, ">");
    }

  strcpy (text, temp);

  return (text);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTReferenceCName (DbTReference This)
/*---------------------------------------------------------------------*/
{
  return ("DbReference");
}

/*- procedure ---------------------------------------------------------*/
static char* DbTReferenceCppName (DbTReference This)
/*---------------------------------------------------------------------*/
{
  static char text[256];
  char temp[240];
    
  char* name;
    
  strcpy (temp, "DbReference");

  if (This->element != NULL)
    {
      strcat (temp, "<");
      name = DbTypeCppName (This->element);
      if (strlen (name) > sizeof(temp))
        {
            /* strcat (temp, "Strange type name"); */
        }
      else
        {
          strcat (temp, name);
        }
      strcat (temp, "> ");
    }

  strcpy (text, temp);

  return (text);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTReferenceSubTypeName (DbTReference This)
/*---------------------------------------------------------------------*/
{
  return (DbTypeCName (This->element));
}

/*- procedure ---------------------------------------------------------*/
static int DbTReferenceSize (DbTReference This)
/*---------------------------------------------------------------------*/
{
  return (sizeof (DbReference));
}

/*- procedure ---------------------------------------------------------*/
static int DbTReferenceAlignment (DbTReference This)
/*---------------------------------------------------------------------*/
{
  DbBuildAlignStruct (DbReference);

  return (DbComputeAlign);
}

/*- procedure ---------------------------------------------------------*/
static void* DbTReferenceInstanciate (DbTReference This)
/*---------------------------------------------------------------------*/
{
  return (DbReferenceNew ());
}

/*- procedure ---------------------------------------------------------*/
static void DbTReferenceInit (DbTReference This, void* address)
/*---------------------------------------------------------------------*/
{
  DbReference* reference = (DbReference*) address;

  DbReferenceInit (reference);
}

/*- procedure ---------------------------------------------------------*/
static void DbTReferenceErase (DbTReference This, void* address)
/*---------------------------------------------------------------------*/
{
  DbReference* reference = (DbReference*) address;

  DbReferenceErase (reference);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTReferenceOutput (DbTReference type, void* address)
/*---------------------------------------------------------------------*/
{
  DbReference* value = (DbReference*) address;

  return (DbReferenceGetName (value));
}

/*---------------------------------------------------------------------*/
static int DbTReferenceIsSame (DbTReference This, DbTReference other)
/*---------------------------------------------------------------------*/
{
  return (DbTypeIsSame (This->element, other->element));
}

/*---------------------------------------------------------------------*/
static DbType DbTReferenceUsedStructure (DbTReference This)
/*---------------------------------------------------------------------*/
{
  return (DbTypeGetUsedStructure (This->element));
}

/*---------------------------------------------------------------------*/
/*                                                                     */
/* Vector  management.                                                 */
/*                                                                     */
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
typedef struct _DbTVectorIteratorRec* DbTVectorIterator;
typedef struct _DbVectorIteratorRec* DbVectorIterator;
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
typedef struct _DbVectorIteratorRec 
{
  DbTypeIteratorRec base;

  DbVector* object;
  int index;
} DbVectorIteratorRec;

static void DbVectorIteratorClear (DbVectorIterator This);
static void DbVectorIteratorFirst (DbVectorIterator This);
static void DbVectorIteratorForth (DbVectorIterator This);
static void DbVectorIteratorExtend (DbVectorIterator This);
static DbTypeIterator DbVectorIteratorDown (DbVectorIterator This);

static DbTypeIteratorClassRec DbVectorIteratorClass = {
  (DbIRoutine) DbVectorIteratorClear,
  (DbIRoutine) DbVectorIteratorFirst,
  (DbIRoutine) DbVectorIteratorForth,
  (DbIMoveRoutine) NULL,
  (DbIRoutine) DbVectorIteratorExtend,
  (DbIteratorIRoutine) DbVectorIteratorDown,
  (DbIteratorIRoutine) NULL,
  (DbIKeyRoutine) NULL
};
/*---------------------------------------------------------------------*/

/*- procedure ---------------------------------------------------------*/
static DbTypeIterator DbVectorIteratorNew (DbTypeIterator parent, 
                                           DbType type, 
                                           IteratorMode mode, 
                                           DbVector* object)
/*---------------------------------------------------------------------*/
{
  DbVectorIterator This;

  if (type == NULL) return (NULL);
  if (object == NULL) return (NULL);

  This = (DbVectorIterator) malloc (sizeof (DbVectorIteratorRec));
  DbTypeIteratorCount++;

  This->base.classDef = &DbVectorIteratorClass;

  This->base.parent    = parent;
  This->base.type      = type;
  This->base.address   = NULL;
  This->base.mode      = mode;
  This->base.finished  = 1;

  This->object = object;
  This->index  = 0;

  DbVectorIteratorFirst (This);

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Create VectorIterator (type %s) #%d\n", 
               DbTypeName (This->base.type), DbTypeIteratorCount);
    }

  return ((DbTypeIterator) This);
}

/*- procedure ---------------------------------------------------------*/
static void DbVectorIteratorClear (DbVectorIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  This->object = NULL;
}

/*- procedure ---------------------------------------------------------*/
static void DbVectorIteratorFirst (DbVectorIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  This->base.finished = 1;

  if (This->object == NULL) return;

  This->base.finished = 0;
  This->base.address = This->object->contents;

  This->index = 0;

  if (This->index >= This->object->size)
    {
      This->base.address = NULL;

      if (This->base.mode != write_only)
        {
          This->base.finished = 1;
        }
    }
}

/*- procedure ---------------------------------------------------------*/
static void DbVectorIteratorForth (DbVectorIterator This)
/*---------------------------------------------------------------------*/
{
  char* address;

  if (This == NULL) return;
  if (This->base.finished) return;

  This->index++;

  if (This->index >= This->object->size)
    {
      if (This->base.mode != write_only)
        {
          This->base.address = NULL;
          This->base.finished = 1;
          return;
        }
      else
        {
          This->base.address = NULL;
        }
    }
  else
    {
      This->base.address = DbVectorGetAddress (This->object, This->index);
    }
}

/*- procedure ---------------------------------------------------------*/
static void DbVectorIteratorExtend (DbVectorIterator This)
/*---------------------------------------------------------------------*/
{
  DbVectorExtend (This->object, This->object->size + 1);

  This->base.address = DbVectorGetAddress (This->object, This->index);
}

/*---------------------------------------------------------------------*/
static DbTypeIterator DbVectorIteratorDown (DbVectorIterator This)
/*---------------------------------------------------------------------*/
{
  DbTypeIterator it;
  DbTypeIterator sub;

  if (This == NULL) return (NULL);
  if (This->base.finished) return (NULL);

  switch (This->base.mode)
    {
      case read_only :
        it = DbTypeGiveInputIterator (This->base.type, 
                                      This->base.address);
        break;
      case visit_only :
        it = DbTypeGiveVisitIterator (This->base.type);
        break;
      case write_only :
        if (This->base.address == NULL)
          {
            DbVectorIteratorExtend (This);
          }
        
        it = DbTypeGiveOutputIterator (This->base.type, 
                                       This->base.address);
    }

  sub = DbTypeIteratorDown (it);

  DbTypeIteratorSetParent (sub, (DbTypeIterator) This);

  DbTypeIteratorUp (it);

  return (sub);
}

/*---------------------------------------------------------------------*/
typedef struct _DbTVectorIteratorRec 
{
  DbTypeIteratorRec base;
} DbTVectorIteratorRec;

static DbTypeIterator DbTVectorIteratorNew (DbTypeIterator parent, 
                                            DbType type,
                                            IteratorMode mode,
                                            DbVector* object);
static void DbTVectorIteratorClear (DbTVectorIterator This);
static void DbTVectorIteratorFirst (DbTVectorIterator This);
static void DbTVectorIteratorForth (DbTVectorIterator This);
static DbTypeIterator DbTVectorIteratorDown (DbTVectorIterator This);

static DbTypeIteratorClassRec DbTVectorIteratorClass = {
  (DbIRoutine) DbTVectorIteratorClear,
  (DbIRoutine) DbTVectorIteratorFirst,
  (DbIRoutine) DbTVectorIteratorForth,
  (DbIMoveRoutine) NULL,
  (DbIRoutine) NULL,
  (DbIteratorIRoutine) DbTVectorIteratorDown,
  (DbIteratorIRoutine) NULL,
  (DbIKeyRoutine) NULL
};
/*---------------------------------------------------------------------*/

/*- procedure ---------------------------------------------------------*/
static DbTypeIterator DbTVectorIteratorNew (DbTypeIterator parent, 
                                            DbType type, 
                                            IteratorMode mode, 
                                            DbVector* object)
/*---------------------------------------------------------------------*/
{
  DbTVectorIterator This;

  if (type == NULL) return (NULL);
  if (object == NULL) return (NULL);

  This = (DbTVectorIterator) malloc (sizeof (DbTVectorIteratorRec));
  DbTypeIteratorCount++;

  This->base.classDef = &DbTVectorIteratorClass;

  This->base.parent    = parent;
  This->base.type      = type;
  This->base.address   = object;
  This->base.mode      = mode;
  This->base.finished  = 1;

  DbTVectorIteratorFirst (This);

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Create TVectorIterator (type %s) #%d\n", 
               DbTypeName (This->base.type), DbTypeIteratorCount);
    }

  return ((DbTypeIterator) This);
}

/*- procedure ---------------------------------------------------------*/
static void DbTVectorIteratorClear (DbTVectorIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;
}

/*- procedure ---------------------------------------------------------*/
static void DbTVectorIteratorFirst (DbTVectorIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  This->base.finished = 0;
}

/*- procedure ---------------------------------------------------------*/
static void DbTVectorIteratorForth (DbTVectorIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;
  if (This->base.finished) return;

  This->base.finished = 1;
}

/*- procedure ---------------------------------------------------------*/
static DbTypeIterator DbTVectorIteratorDown (DbTVectorIterator This)
/*---------------------------------------------------------------------*/
{
  DbVector* object;
  DbTypeIterator sub;

  if (This == NULL) return (NULL);
  if (This->base.finished) return (NULL);

  object = (DbVector*) This->base.address;

  This->base.finished = 0;

  sub = DbVectorIteratorNew ((DbTypeIterator) This, 
                             DbVectorGetElementType (object), 
                             This->base.mode, object);

  return (sub);
}

/*---------------------------------------------------------------------*/
typedef struct _DbTVectorRec {
  DbTypeRec base;
  DbType    element;
} DbTVectorRec;

static void DbTVectorClear (DbTVector This);
static char* DbTVectorName (DbTVector This);
static char* DbTVectorCName (DbTVector This);
static char* DbTVectorCppName (DbTVector This);
static char* DbTVectorSubTypeName (DbTVector This);
static int DbTVectorSize (DbTVector This);
static int DbTVectorAlignment (DbTVector This);
static void* DbTVectorInstanciate (DbTVector This);
static void DbTVectorInit (DbTVector This, void* address);
static void DbTVectorErase (DbTVector This, void* address);
static DbTypeIterator DbTVectorGiveIterator (DbTVector This, 
                                             IteratorMode mode, 
                                             void* address);
static int DbTVectorIsSame (DbTVector This, DbTVector other);
static DbType DbTVectorUsedStructure (DbTVector This);

static DbTypeClassRec DbTVectorClass = {
  "DbTVector",

  (DbRoutine) DbTVectorClear,
  (DbStringRoutine) DbTVectorName,
  (DbStringRoutine) DbTVectorCName,
  (DbStringRoutine) DbTVectorCppName,
  (DbStringRoutine) DbTVectorSubTypeName,
  (DbIntRoutine) DbTVectorSize,
  (DbIntRoutine) DbTVectorAlignment,
  (DbPtrRoutine) DbTVectorInstanciate,
  (DbInitRoutine) DbTVectorInit,
  (DbInitRoutine) DbTVectorErase,
  (DbIteratorRoutine) DbTVectorGiveIterator,
  (DbOutputRoutine) NULL,
  (DbCompareRoutine) DbTVectorIsSame,
  (DbUsedStructureRoutine) DbTVectorUsedStructure
};
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
static CLList Vectors = NULL;
/*---------------------------------------------------------------------*/

/*- procedure ---------------------------------------------------------*/
static void DbTVectorCleanup ()
/*---------------------------------------------------------------------*/
{
  CLListErase (Vectors);
  GCReferenceClear (Vectors);

  GCCleanup ();
}

/*- procedure ---------------------------------------------------------*/
DbType DbTVectorFind (DbType element)
/*---------------------------------------------------------------------*/
{
  CIterator it;
  DbType result = NULL;

  if (Vectors == NULL) return (NULL);

  it = CLListGiveIterator (Vectors);
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbTVector v;

      v = (DbTVector) CIteratorItem (it);
      if (v != NULL)
        {
          if (v->element == element)
            {
              result = (DbType) v;
              break;
            }
        }
    }
  CIteratorDelete (it);
      
  return (result);
}

/*- procedure ---------------------------------------------------------*/
DbType DbTVectorNew (DbType element)
/*---------------------------------------------------------------------*/
{
  DbTVector This;

  if (element == NULL) return (NULL);

  This = (DbTVector) DbTVectorFind (element);
  if (This != NULL) return ((DbType) This);

  DbTypeStartup ();

  This = (DbTVector) malloc (sizeof (DbTVectorRec));
  DbTypeCount++;

  This->base.classDef = &DbTVectorClass;
    
  This->element  = NULL;

  GCObjectNew (This, DbTypeGCClass);

  GCReferenceAssignIn (This->element, This, element);

  if (Vectors == NULL)
    {
      GCReferenceAssign (Vectors, CLListNew ());
    }

  CLListAdd (Vectors, This);

  return ((DbType) This);
}

/*- procedure ---------------------------------------------------------*/
static void DbTVectorClear (DbTVector This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  GCReferenceClear (This->element);
}

/*- procedure ---------------------------------------------------------*/
static void DbTVectorUninstall (DbTVector This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  if (DbTVectorFind (This->element) == NULL) return;
  CLListRemove (Vectors, This);

  DbTypeUninstall ((DbType) This);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTVectorName (DbTVector This)
/*---------------------------------------------------------------------*/
{
  static char text[256];
  char temp[240];
    
  char* name;
    
  name = DbTypeName (This->element);
  if (strlen (name) > sizeof(temp))
    {
      strcpy (temp, "Strange type name");
      temp[0] = 0;
    }
  else
    {
      strcpy (temp, name);
    }

  strcpy (text, temp);
  sprintf (&text[strlen(text)], "[]");

  return (text);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTVectorCName (DbTVector This)
/*---------------------------------------------------------------------*/
{
  return ("DbVector");
}

/*- procedure ---------------------------------------------------------*/
static char* DbTVectorCppName (DbTVector This)
/*---------------------------------------------------------------------*/
{
  static char text[256];
  char temp[240];
    
  char* name;
    
  strcpy (temp, "DbVector");

  if (This->element != NULL)
    {
      strcat (temp, "<");
      name = DbTypeCppName (This->element);
      if (strlen (name) > sizeof(temp))
        {
            /* strcat (temp, "Strange type name"); */
        }
      else
        {
          strcat (temp, name);
        }
      strcat (temp, "> ");
    }

  strcpy (text, temp);

  return (text);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTVectorSubTypeName (DbTVector This)
/*---------------------------------------------------------------------*/
{
  return (DbTypeCName (This->element));
}

/*- procedure ---------------------------------------------------------*/
static int DbTVectorSize (DbTVector This)
/*---------------------------------------------------------------------*/
{
  return (sizeof (DbVector));
}

/*- procedure ---------------------------------------------------------*/
static int DbTVectorAlignment (DbTVector This)
/*---------------------------------------------------------------------*/
{
  DbBuildAlignStruct (DbVector*);

  return (DbComputeAlign);
}

/*- procedure ---------------------------------------------------------*/
static void* DbTVectorInstanciate (DbTVector This)
/*---------------------------------------------------------------------*/
{
  return (DbVectorNew (This->element));
}

/*- procedure ---------------------------------------------------------*/
static void DbTVectorInit (DbTVector This, void* address)
/*---------------------------------------------------------------------*/
{
  DbVector* vector = (DbVector*) address;

  DbVectorInit (vector, This->element);
}

/*- procedure ---------------------------------------------------------*/
static void DbTVectorErase (DbTVector This, void* address)
/*---------------------------------------------------------------------*/
{
  DbVector* vector = (DbVector*) address;

  DbVectorErase (vector);
}

/*- procedure ---------------------------------------------------------*/
static DbTypeIterator DbTVectorGiveIterator (DbTVector This, 
                                             IteratorMode mode, 
                                             void* address)
/*---------------------------------------------------------------------*/
{
  DbTypeIterator it;

  it =  DbTVectorIteratorNew (NULL, (DbType) This, mode, 
                              (DbVector*) address);

  return (it);
}

/*---------------------------------------------------------------------*/
static int DbTVectorIsSame (DbTVector This, DbTVector other)
/*---------------------------------------------------------------------*/
{
  return (DbTypeIsSame (This->element, other->element));
}

/*---------------------------------------------------------------------*/
static DbType DbTVectorUsedStructure (DbTVector This)
/*---------------------------------------------------------------------*/
{
  return (DbTypeGetUsedStructure (This->element));
}

/*---------------------------------------------------------------------*/
/*                                                                     */
/* Matrix  management.                                                 */
/*                                                                     */
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
typedef struct _DbRowIteratorRec* DbRowIterator;
typedef struct _DbColumnIteratorRec* DbColumnIterator;
typedef struct _DbTMatrixIteratorRec* DbTMatrixIterator;
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
typedef struct _DbColumnIteratorRec 
{
  DbTypeIteratorRec base;

  void* row_contents;
  int column;
  int columns;
} DbColumnIteratorRec;

static void DbColumnIteratorClear (DbColumnIterator This);
static void DbColumnIteratorFirst (DbColumnIterator This);
static void DbColumnIteratorForth (DbColumnIterator This);
static DbTypeIterator DbColumnIteratorDown (DbColumnIterator This);

static DbTypeIteratorClassRec DbColumnIteratorClass = {
  (DbIRoutine) DbColumnIteratorClear,
  (DbIRoutine) DbColumnIteratorFirst,
  (DbIRoutine) DbColumnIteratorForth,
  (DbIMoveRoutine) NULL,
  (DbIRoutine) NULL,
  (DbIteratorIRoutine) DbColumnIteratorDown,
  (DbIteratorIRoutine) NULL,
  (DbIKeyRoutine) NULL
};
/*---------------------------------------------------------------------*/

/*- procedure ---------------------------------------------------------*/
static DbTypeIterator DbColumnIteratorNew (DbTypeIterator parent, 
                                           DbType type, 
                                           IteratorMode mode, 
                                           void* address,
                                           int columns)
/*---------------------------------------------------------------------*/
{
  DbColumnIterator This;

  if (type == NULL) return (NULL);
  if (columns <= 0) return (NULL);

  This = (DbColumnIterator) malloc (sizeof (DbColumnIteratorRec));
  DbTypeIteratorCount++;

  This->base.classDef = &DbColumnIteratorClass;

  This->base.parent    = parent;
  This->base.type      = type;
  This->base.address   = NULL;
  This->base.mode      = mode;
  This->base.finished  = 1;

  This->row_contents = address;
  This->column  = 0;
  This->columns = columns;

  DbColumnIteratorFirst (This);

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Create ColumnIterator (type %s) #%d\n", 
               DbTypeName (This->base.type), DbTypeIteratorCount);
    }

  return ((DbTypeIterator) This);
}

/*- procedure ---------------------------------------------------------*/
static void DbColumnIteratorClear (DbColumnIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  This->row_contents = NULL;
}

/*- procedure ---------------------------------------------------------*/
static void DbColumnIteratorFirst (DbColumnIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  This->base.finished = 1;

  if (This->row_contents == NULL) return;

  This->base.finished = 0;
  This->base.address = This->row_contents;

  This->column = 0;

  if (This->column >= This->columns)
    {
      This->base.address = NULL;
      This->base.finished = 1;
    }
}

/*- procedure ---------------------------------------------------------*/
static void DbColumnIteratorForth (DbColumnIterator This)
/*---------------------------------------------------------------------*/
{
  char* address;

  if (This == NULL) return;
  if (This->base.finished) return;

  This->column++;

  if (This->column >= This->columns)
    {
      This->base.address = NULL;
      This->base.finished = 1;
      return;
    }

  address = (char*) This->row_contents;
  address += DbTypeSize (This->base.type) * This->column;

  This->base.address = address;
}

/*---------------------------------------------------------------------*/
static DbTypeIterator DbColumnIteratorDown (DbColumnIterator This)
/*---------------------------------------------------------------------*/
{
  DbTypeIterator it;
  DbTypeIterator sub;

  if (This == NULL) return (NULL);
  if (This->base.finished) return (NULL);

  switch (This->base.mode)
    {
      case read_only :
        it = DbTypeGiveInputIterator (This->base.type, 
                                      This->base.address);
        break;
      case visit_only :
        it = DbTypeGiveVisitIterator (This->base.type);
        break;
      case write_only :
        it = DbTypeGiveOutputIterator (This->base.type, 
                                       This->base.address);
        break;
    }

  sub = DbTypeIteratorDown (it);

  DbTypeIteratorSetParent (sub, (DbTypeIterator) This);

  DbTypeIteratorUp (it);

  return (sub);
}

/*---------------------------------------------------------------------*/
typedef struct _DbRowIteratorRec 
{
  DbTypeIteratorRec base;

  DbMatrix* object;
  int row;
} DbRowIteratorRec;

static void DbRowIteratorClear (DbRowIterator This);
static void DbRowIteratorFirst (DbRowIterator This);
static void DbRowIteratorForth (DbRowIterator This);
static DbTypeIterator DbRowIteratorDown (DbRowIterator This);

static DbTypeIteratorClassRec DbRowIteratorClass = {
  (DbIRoutine) DbRowIteratorClear,
  (DbIRoutine) DbRowIteratorFirst,
  (DbIRoutine) DbRowIteratorForth,
  (DbIMoveRoutine) NULL,
  (DbIRoutine) NULL,
  (DbIteratorIRoutine) DbRowIteratorDown,
  (DbIteratorIRoutine) NULL,
  (DbIKeyRoutine) NULL
};
/*---------------------------------------------------------------------*/

/*- procedure ---------------------------------------------------------*/
static DbTypeIterator DbRowIteratorNew (DbTypeIterator parent, 
                                        DbType type, 
                                        IteratorMode mode, 
                                        DbMatrix* object)
/*---------------------------------------------------------------------*/
{
  DbRowIterator This;

  if (type == NULL) return (NULL);
  if (object == NULL) return (NULL);
  if (object->rows == 0) return (NULL);

  This = (DbRowIterator) malloc (sizeof (DbRowIteratorRec));
  DbTypeIteratorCount++;

  This->base.classDef = &DbRowIteratorClass;

  This->base.parent    = parent;
  This->base.type      = type;
  This->base.address   = NULL;
  This->base.mode      = mode;
  This->base.finished  = 1;

  This->object = object;
  This->row    = 0;

  DbRowIteratorFirst (This);

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Create RowIterator (type %s) #%d\n", 
               DbTypeName (This->base.type), DbTypeIteratorCount);
    }

  return ((DbTypeIterator) This);
}

/*- procedure ---------------------------------------------------------*/
static void DbRowIteratorClear (DbRowIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  This->object = NULL;
}

/*- procedure ---------------------------------------------------------*/
static void DbRowIteratorFirst (DbRowIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  This->base.finished = 1;

  if (This->object == NULL) return;

  This->base.finished = 0;
  This->base.address = This->object->contents;

  This->row = 0;

  if (This->row >= This->object->rows)
    {
      This->base.address = NULL;
      This->base.finished = 1;
    }
}

/*- procedure ---------------------------------------------------------*/
static void DbRowIteratorForth (DbRowIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;
  if (This->base.finished) return;

  This->row++;

  if (This->row >= This->object->rows)
    {
      This->base.address = NULL;
      This->base.finished = 1;
      return;
    }

  This->base.address = DbMatrixGetAddress (This->object, This->row, 0);
}

/*- procedure ---------------------------------------------------------*/
static DbTypeIterator DbRowIteratorDown (DbRowIterator This)
/*---------------------------------------------------------------------*/
{
  DbTypeIterator sub;

  if (This == NULL) return (NULL);
  if (This->base.finished) return (NULL);

  This->base.finished = 0;

  sub = DbColumnIteratorNew ((DbTypeIterator) This, 
                             DbMatrixGetElementType (This->object), 
                             This->base.mode, This->base.address,
                             This->object->columns);

  return (sub);
}

/*---------------------------------------------------------------------*/
typedef struct _DbTMatrixIteratorRec 
{
  DbTypeIteratorRec base;
} DbTMatrixIteratorRec;

static DbTypeIterator DbTMatrixIteratorNew (DbTypeIterator parent, 
                                            DbType type, 
                                            IteratorMode mode, 
                                            void* address);
static void DbTMatrixIteratorClear (DbTMatrixIterator This);
static void DbTMatrixIteratorFirst (DbTMatrixIterator This);
static void DbTMatrixIteratorForth (DbTMatrixIterator This);
static DbTypeIterator DbTMatrixIteratorDown (DbTMatrixIterator This);

static DbTypeIteratorClassRec DbTMatrixIteratorClass = {
  (DbIRoutine) DbTMatrixIteratorClear,
  (DbIRoutine) DbTMatrixIteratorFirst,
  (DbIRoutine) DbTMatrixIteratorForth,
  (DbIMoveRoutine) NULL,
  (DbIRoutine) NULL,
  (DbIteratorIRoutine) DbTMatrixIteratorDown,
  (DbIteratorIRoutine) NULL,
  (DbIKeyRoutine) NULL
};
/*---------------------------------------------------------------------*/

/*- procedure ---------------------------------------------------------*/
static DbTypeIterator DbTMatrixIteratorNew (DbTypeIterator parent, 
                                            DbType type, 
                                            IteratorMode mode, 
                                            void* address)
/*---------------------------------------------------------------------*/
{
  DbTMatrixIterator This;

  if (type == NULL) return (NULL);

  This = (DbTMatrixIterator) malloc (sizeof (DbTMatrixIteratorRec));
  DbTypeIteratorCount++;

  This->base.classDef = &DbTMatrixIteratorClass;

  This->base.parent   = parent;
  This->base.type     = type;
  This->base.address  = address;
  This->base.mode     = mode;
  This->base.finished = 1;

  DbTMatrixIteratorFirst (This);
 
  if (Debug)
    {
      fprintf (stderr,
               "DbType> Create TMatrixIterator (type %s) #%d\n", 
               DbTypeName (This->base.type), DbTypeIteratorCount);
    }

  return ((DbTypeIterator) This);
}

/*- procedure ---------------------------------------------------------*/
static void DbTMatrixIteratorClear (DbTMatrixIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;
}

/*- procedure ---------------------------------------------------------*/
static void DbTMatrixIteratorFirst (DbTMatrixIterator This)
/*---------------------------------------------------------------------*/
{
  DbMatrix* object;

  if (This == NULL) return;

  This->base.finished = 1;

  object = (DbMatrix*) This->base.address;
  if (object == NULL) return;

  if ((object->rows == 0) || (object->columns == 0)) return;

  This->base.finished = 0;
}

/*- procedure ---------------------------------------------------------*/
static void DbTMatrixIteratorForth (DbTMatrixIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;
  if (This->base.finished) return;

  This->base.finished = 1;
}

/*- procedure ---------------------------------------------------------*/
static DbTypeIterator DbTMatrixIteratorDown (DbTMatrixIterator This)
/*---------------------------------------------------------------------*/
{
  DbMatrix* object;
  DbTypeIterator sub;

  if (This == NULL) return (NULL);
  if (This->base.finished) return (NULL);

  object = (DbMatrix*) This->base.address;

  This->base.finished = 0;

  sub = DbRowIteratorNew ((DbTypeIterator) This, 
                          DbMatrixGetElementType (object), 
                          This->base.mode, object);

  return (sub);
}

/*---------------------------------------------------------------------*/
typedef struct _DbTMatrixRec {
  DbTypeRec base;
  DbType    element;
  int       rows;
  int       columns;
} DbTMatrixRec;

static void DbTMatrixClear (DbTMatrix This);
static char* DbTMatrixName (DbTMatrix This);
static char* DbTMatrixCName (DbTMatrix This);
static char* DbTMatrixCppName (DbTMatrix This);
static char* DbTMatrixSubTypeName (DbTMatrix This);
static int DbTMatrixSize (DbTMatrix This);
static int DbTMatrixAlignment (DbTMatrix This);
static void* DbTMatrixInstanciate (DbTMatrix This);
static void DbTMatrixInit (DbTMatrix This, void* address);
static void DbTMatrixErase (DbTMatrix This, void* address);
static DbTypeIterator DbTMatrixGiveIterator (DbTMatrix This, 
                                             IteratorMode mode, 
                                             void* address);
static int DbTMatrixIsSame (DbTMatrix This, DbTMatrix other);
static DbType DbTMatrixUsedStructure (DbTMatrix This);

static DbTypeClassRec DbTMatrixClass = {
  "DbTMatrix",

  (DbRoutine) DbTMatrixClear,
  (DbStringRoutine) DbTMatrixName,
  (DbStringRoutine) DbTMatrixCName,
  (DbStringRoutine) DbTMatrixCppName,
  (DbStringRoutine) DbTMatrixSubTypeName,
  (DbIntRoutine) DbTMatrixSize,
  (DbIntRoutine) DbTMatrixAlignment,
  (DbPtrRoutine) DbTMatrixInstanciate,
  (DbInitRoutine) DbTMatrixInit,
  (DbInitRoutine) DbTMatrixErase,
  (DbIteratorRoutine) DbTMatrixGiveIterator,
  (DbOutputRoutine) NULL,
  (DbCompareRoutine) DbTMatrixIsSame,
  (DbUsedStructureRoutine) DbTMatrixUsedStructure
};
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
static CLList Matrices = NULL;
/*---------------------------------------------------------------------*/

/*- procedure ---------------------------------------------------------*/
static void DbTMatrixCleanup ()
/*---------------------------------------------------------------------*/
{
  CLListErase (Matrices);
  GCReferenceClear (Matrices);

  GCCleanup ();
}

/*- procedure ---------------------------------------------------------*/
DbType DbTMatrixFind (DbType element, int rows, int columns)
/*---------------------------------------------------------------------*/
{
  CIterator it;
  DbType result = NULL;

  if (Matrices == NULL) return (NULL);

  it = CLListGiveIterator (Matrices);
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbTMatrix m;

      m = (DbTMatrix) CIteratorItem (it);
      if (m != NULL)
        {
          if ((m->element == element) &&
              (m->rows == rows) &&
              (m->columns == columns))
            {
              result = (DbType) m;
              break;
            }
        }
    }
  CIteratorDelete (it);
      
  return (result);
}

/*- procedure ---------------------------------------------------------*/
DbType DbTMatrixNew (DbType element, int rows, int columns)
/*---------------------------------------------------------------------*/
{
  DbTMatrix This;

  if (element == NULL) return (NULL);

  This = (DbTMatrix) DbTMatrixFind (element, rows, columns);
  if (This != NULL) return ((DbType) This);

  DbTypeStartup ();

  This = (DbTMatrix) malloc (sizeof (DbTMatrixRec));
  DbTypeCount++;

  This->base.classDef = &DbTMatrixClass;
    
  This->element  = NULL;
  This->rows     = rows;
  This->columns  = columns;

  GCObjectNew (This, DbTypeGCClass);

  GCReferenceAssignIn (This->element, This, element);

  return ((DbType) This);
}

/*- procedure ---------------------------------------------------------*/
static void DbTMatrixUninstall (DbTMatrix This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  if (DbTMatrixFind (This->element, This->rows, This->columns) == NULL) return;
  CLListRemove (Matrices, This);

  DbTypeUninstall ((DbType) This);
}

/*- procedure ---------------------------------------------------------*/
static void DbTMatrixClear (DbTMatrix This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  GCReferenceClear (This->element);

  This->rows = 0;
  This->columns = 0;
}

/*- procedure ---------------------------------------------------------*/
static char* DbTMatrixName (DbTMatrix This)
/*---------------------------------------------------------------------*/
{
  static char text[256];
  char temp[240];
    
  char* name;
    
  name = DbTypeName (This->element);
  if (strlen (name) > sizeof(temp))
    strcpy (temp, "Strange type name");
  else
    strcpy (temp, DbTypeName (This->element));

  strcpy (text, temp);
  sprintf (&text[strlen(text)], "[%d][%d]", This->rows, This->columns);

  return (text);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTMatrixCName (DbTMatrix This)
/*---------------------------------------------------------------------*/
{
  return ("DbMatrix");
}

/*- procedure ---------------------------------------------------------*/
static char* DbTMatrixCppName (DbTMatrix This)
/*---------------------------------------------------------------------*/
{
  static char text[256];
  char temp[240];
    
  char* name;
    
  strcpy (temp, "DbMatrix");

  if (This->element != NULL)
    {
      strcat (temp, "<");
      name = DbTypeCppName (This->element);
      if (strlen (name) > sizeof(temp))
        {
            /* strcat (temp, "Strange type name"); */
        }
      else
        {
          strcat (temp, name);
        }
      strcat (temp, "> ");
    }

  strcpy (text, temp);

  return (text);
}

/*- procedure ---------------------------------------------------------*/
static char* DbTMatrixSubTypeName (DbTMatrix This)
/*---------------------------------------------------------------------*/
{
  return (DbTypeCName (This->element));
}

/*- procedure ---------------------------------------------------------*/
static int DbTMatrixSize (DbTMatrix This)
/*---------------------------------------------------------------------*/
{
  return (sizeof (DbMatrix));
}

/*- procedure ---------------------------------------------------------*/
static int DbTMatrixAlignment (DbTMatrix This)
/*---------------------------------------------------------------------*/
{
  DbBuildAlignStruct (DbMatrix);

  return (DbComputeAlign);
}

/*- procedure ---------------------------------------------------------*/
static void* DbTMatrixInstanciate (DbTMatrix This)
/*---------------------------------------------------------------------*/
{
  return (DbMatrixNew (This->element, This->rows, This->columns));
}

/*- procedure ---------------------------------------------------------*/
static void DbTMatrixInit (DbTMatrix This, void* address)
/*---------------------------------------------------------------------*/
{
  DbMatrix* matrix = (DbMatrix*) address;

  DbMatrixInit (matrix, This->element, This->rows, This->columns);
}

/*- procedure ---------------------------------------------------------*/
static void DbTMatrixErase (DbTMatrix This, void* address)
/*---------------------------------------------------------------------*/
{
  DbMatrix* matrix = (DbMatrix*) address;

  DbMatrixErase (matrix);
}

/*- procedure ---------------------------------------------------------*/
static DbTypeIterator DbTMatrixGiveIterator (DbTMatrix This, 
                                             IteratorMode mode, 
                                             void* address)
/*---------------------------------------------------------------------*/
{
  DbTypeIterator it;

  it =  DbTMatrixIteratorNew (NULL, (DbType) This, mode, address);

  return (it);
}

/*---------------------------------------------------------------------*/
static int DbTMatrixIsSame (DbTMatrix This, DbTMatrix other)
/*---------------------------------------------------------------------*/
{
  return (DbTypeIsSame (This->element, other->element));
}

/*---------------------------------------------------------------------*/
static DbType DbTMatrixUsedStructure (DbTMatrix This)
/*---------------------------------------------------------------------*/
{
  return (DbTypeGetUsedStructure (This->element));
}

/*---------------------------------------------------------------------*/
/*                                                                     */
/*  Field     management.                                              */
/*                                                                     */
/*---------------------------------------------------------------------*/

static void DbFieldDelete (DbField This);

typedef struct _DbFieldRec {
  DbType type;
  char* name;
  int offset;
} DbFieldRec;

/*- procedure ---------------------------------------------------------*/
DbField DbFieldNew (DbType type, char* name)
/*---------------------------------------------------------------------*/
{
  DbField This;

  This = (DbField) malloc (sizeof (DbFieldRec));
  DbFieldCount++;

  This->name = (char*) malloc (strlen(name) + 1);
  DbFieldNameCount++;
  strcpy (This->name, name);
  This->type = NULL;
  This->offset = 0;

  GCObjectNew (This, DbFieldGCClass);

  GCReferenceAssignIn (This->type, This, type);

  return (This);
}

/*- procedure ---------------------------------------------------------*/
static void DbFieldErase (DbField This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  GCReferenceClear (This->type);
}

/*- procedure ---------------------------------------------------------*/
static void DbFieldDelete (DbField This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  if (This->name)
  {
    free (This->name);
    DbFieldNameCount--;

    This->name = 0;
  }
  GCReferenceClear (This->type);

  free (This);
  DbFieldCount--;
}

/*- procedure ---------------------------------------------------------*/
DbType DbFieldType (DbField This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (NULL);
  
  return (This->type);
}

/*- procedure ---------------------------------------------------------*/
char* DbFieldGetName (DbField This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (NULL);
  
  return (This->name);
}

/*- procedure ---------------------------------------------------------*/
int DbFieldIsSame (DbField This, DbField other)
/*---------------------------------------------------------------------*/
{
  if (This == other) return (1);
  if (This == NULL) return (0);
  if (other == NULL) return (0);

  if (strcmp (This->name, other->name) != 0) return (0);
  
  return (DbTypeIsSame (This->type, other->type));
}

/*---------------------------------------------------------------------*/
static int DbFieldListCompare (CLList fields1, CLList fields2)
/*---------------------------------------------------------------------*/
{
  int result = 1;
  CIterator it1;
  CIterator it2;

  it1 = CLListGiveIterator (fields1);
  it2 = CLListGiveIterator (fields2);

  while (!CIteratorFinished (it1) &&
         !CIteratorFinished (it2))
    {
      DbField field1;
      DbField field2;
      
      field1 = (DbField) CIteratorItem (it1);
      field2 = (DbField) CIteratorItem (it2);

      if (!DbFieldIsSame (field1, field2))
        {
          result = 0;
          break;
        }

      CIteratorForth (it1);
      CIteratorForth (it2);
    }
  
  CIteratorDelete (it1);
  CIteratorDelete (it2);

  return (result);
}

/*---------------------------------------------------------------------*/
/*                                                                     */
/* Structure  management.                                              */
/*                                                                     */
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
typedef struct _DbStructureIteratorRec* DbStructureIterator;
typedef struct _DbTStructureIteratorRec* DbTStructureIterator;
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
typedef struct _DbTStructureRec {
  DbTypeRec base;
  char* name;
  CLList fields;
  int size;
  int align;

  CHasher used_structures;
} DbTStructureRec;

typedef struct _DbStructureIteratorRec 
{
  DbTypeIteratorRec base;

  DbTStructure structure;
  CIterator field_it;
  DbField field;
  void* object;
} DbStructureIteratorRec;

static DbTypeIterator DbStructureIteratorNew (DbTypeIterator parent, 
                                              DbTStructure type, 
                                              IteratorMode mode, 
                                              void* address);
static void DbStructureIteratorClear (DbStructureIterator This);
static void DbStructureIteratorFirst (DbStructureIterator This);
static void DbStructureIteratorForth (DbStructureIterator This);
static void DbStructureIteratorMoveTo (DbStructureIterator This, char* key);
static DbTypeIterator DbStructureIteratorDown (DbStructureIterator This);
static char* DbStructureIteratorGetKey (DbStructureIterator This);

static DbTypeIteratorClassRec DbStructureIteratorClass = {
  (DbIRoutine) DbStructureIteratorClear,
  (DbIRoutine) DbStructureIteratorFirst,
  (DbIRoutine) DbStructureIteratorForth,
  (DbIMoveRoutine) DbStructureIteratorMoveTo,
  (DbIRoutine) NULL,
  (DbIteratorIRoutine) DbStructureIteratorDown,
  (DbIteratorIRoutine) NULL,
  (DbIKeyRoutine) DbStructureIteratorGetKey
};
/*---------------------------------------------------------------------*/

/*- procedure ---------------------------------------------------------*/
static DbTypeIterator DbStructureIteratorNew (DbTypeIterator parent, 
                                              DbTStructure type, 
                                              IteratorMode mode, 
                                              void* address)
/*---------------------------------------------------------------------*/
{
  DbStructureIterator This;
  void* object = address;

  if (type == NULL) return (NULL);

  This = (DbStructureIterator) malloc (sizeof (DbStructureIteratorRec));
  DbTypeIteratorCount++;

  This->base.classDef = &DbStructureIteratorClass;

  This->base.parent    = parent;
  This->base.type      = NULL;
  This->base.address   = NULL;
  This->base.mode      = mode;
  This->base.finished  = 1;

  This->structure = type;
  This->field_it  = NULL;
  This->field     = NULL;
  This->object    = address;

  DbStructureIteratorFirst (This);

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Create StructureIterator (type %s) #%d\n", 
               DbTypeName (This->base.type), DbTypeIteratorCount);
    }

  return ((DbTypeIterator) This);
}

/*- procedure ---------------------------------------------------------*/
static void DbStructureIteratorClear (DbStructureIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  This->structure = NULL;

  if (This->field_it != NULL)
    {
      CIteratorDelete (This->field_it);
      This->field_it = NULL;
    }

  This->field = NULL;
  This->object = NULL;
}

/*- procedure ---------------------------------------------------------*/
static void DbStructureIteratorFirst (DbStructureIterator This)
/*---------------------------------------------------------------------*/
{
  char* address;
  DbType field_type;
  DbTStructure structure;

  if (This == NULL) return;

  This->base.finished = 1;

  if (This->field_it != NULL)
    {
      CIteratorDelete (This->field_it);
      This->field_it = NULL;
    }

  structure = (DbTStructure) This->structure;

  This->field_it = CLListGiveIterator (structure->fields);

  if (!CIteratorFinished (This->field_it))
    {
      int offset;

      This->field = (DbField) CIteratorItem (This->field_it);
      field_type = This->field->type;

      offset = This->field->offset;

      address = (char*) This->object + offset;

      This->base.type = field_type;
      This->base.address = (void*) address;
      
      This->base.finished = 0;
    }
  else
    {
      CIteratorDelete (This->field_it);
      This->field_it = NULL;
    }
}

/*- procedure ---------------------------------------------------------*/
static void DbStructureIteratorForth (DbStructureIterator This)
/*---------------------------------------------------------------------*/
{
  char* address;
  int offset;
  DbType field_type;
  DbTStructure structure;

  if (This == NULL) return;
  if (This->base.finished) return;

  CIteratorForth (This->field_it);

  if (!CIteratorFinished (This->field_it))
    {
        /* Get the next field */
      This->field = (DbField) CIteratorItem (This->field_it);
      field_type = This->field->type;

      address = (char*) This->object + This->field->offset;

        /* Set the current field to it */
      This->base.type    = field_type;
      This->base.address = (void*) address;
    }
  else
    {
      This->base.finished = 1;
      CIteratorDelete (This->field_it);
      This->field_it = NULL;
    }
}

/*- procedure ---------------------------------------------------------*/
static void DbStructureIteratorMoveTo (DbStructureIterator This, char* key)
/*---------------------------------------------------------------------*/
{
  char* address;
  int offset;
  DbType field_type;
  DbTStructure structure;

  if (This == NULL) return;

  DbStructureIteratorFirst (This);

  for (;!This->base.finished; DbStructureIteratorForth (This))
    {
      if (strcmp (key, This->field->name) == 0)
        {
          This->base.finished = 0;
          break;
        }
    }
}

/*---------------------------------------------------------------------*/
static DbTypeIterator DbStructureIteratorDown (DbStructureIterator This)
/*---------------------------------------------------------------------*/
{
  DbTypeIterator it;
  DbTypeIterator sub;

  if (This == NULL) return (NULL);
  if (This->base.finished) return (NULL);

  switch (This->base.mode)
    {
      case read_only :
        it = DbTypeGiveInputIterator (This->base.type, 
                                      This->base.address);
        break;
      case visit_only :
        it = DbTypeGiveVisitIterator (This->base.type);
        break;
      case write_only :
        it = DbTypeGiveOutputIterator (This->base.type, 
                                       This->base.address);
        break;
    }

  sub = DbTypeIteratorDown (it);

  DbTypeIteratorSetParent (sub, (DbTypeIterator) This);

  DbTypeIteratorUp (it);

  return (sub);
}

/*- procedure ---------------------------------------------------------*/
static char* DbStructureIteratorGetKey (DbStructureIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (NULL);
  if (This->base.finished) return (NULL);

  return (DbFieldGetName (This->field));
}

/*---------------------------------------------------------------------*/
typedef struct _DbTStructureIteratorRec 
{
  DbTypeIteratorRec base;
} DbTStructureIteratorRec;

static DbTypeIterator DbTStructureIteratorNew (DbTypeIterator parent, 
                                               DbTStructure type, 
                                               IteratorMode mode, 
                                               void* object);
static void DbTStructureIteratorClear (DbTStructureIterator This);
static void DbTStructureIteratorFirst (DbTStructureIterator This);
static void DbTStructureIteratorForth (DbTStructureIterator This);
static DbTypeIterator DbTStructureIteratorDown (DbTStructureIterator This);

static DbTypeIteratorClassRec DbTStructureIteratorClass = {
  (DbIRoutine) DbTStructureIteratorClear,
  (DbIRoutine) DbTStructureIteratorFirst,
  (DbIRoutine) DbTStructureIteratorForth,
  (DbIMoveRoutine) NULL,
  (DbIRoutine) NULL,
  (DbIteratorIRoutine) DbTStructureIteratorDown,
  (DbIteratorIRoutine) NULL,
  (DbIKeyRoutine) NULL
};
/*---------------------------------------------------------------------*/

/*- procedure ---------------------------------------------------------*/
static DbTypeIterator DbTStructureIteratorNew (DbTypeIterator parent, 
                                               DbTStructure type, 
                                               IteratorMode mode, 
                                               void* object)
/*---------------------------------------------------------------------*/
{
  DbTStructureIterator This;

  if (type == NULL) return (NULL);
  if (object == NULL) return (NULL);

  This = (DbTStructureIterator) malloc (sizeof (DbTStructureIteratorRec));
  DbTypeIteratorCount++;

  This->base.classDef = &DbTStructureIteratorClass;

  This->base.parent    = parent;
  This->base.type      = (DbType) type;
  This->base.address   = object;
  This->base.mode      = mode;
  This->base.finished  = 1;

  DbTStructureIteratorFirst (This);

  if (Debug)
    {
      fprintf (stderr,
               "DbType> Create TStructureIterator (type %s) #%d\n", 
               DbTypeName (This->base.type), DbTypeIteratorCount);
    }

  return ((DbTypeIterator) This);
}

/*- procedure ---------------------------------------------------------*/
static void DbTStructureIteratorClear (DbTStructureIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;
}

/*- procedure ---------------------------------------------------------*/
static void DbTStructureIteratorFirst (DbTStructureIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  This->base.finished = 0;
}

/*- procedure ---------------------------------------------------------*/
static void DbTStructureIteratorForth (DbTStructureIterator This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return;

  This->base.finished = 1;
}

/*- procedure ---------------------------------------------------------*/
static DbTypeIterator DbTStructureIteratorDown (DbTStructureIterator This)
/*---------------------------------------------------------------------*/
{
  DbTypeIterator sub;

  if (This == NULL) return (NULL);
  if (This->base.finished) return (NULL);

  This->base.finished = 0;

  sub = DbStructureIteratorNew ((DbTypeIterator) This, 
                                (DbTStructure) This->base.type, 
                                This->base.mode, 
                                This->base.address);

  return (sub);
}

/*---------------------------------------------------------------------*/
static void DbTStructureClear (DbTStructure This);
static char* DbTStructureName (DbTStructure This);
static int DbTStructureSize (DbTStructure This);
static int DbTStructureAlignment (DbTStructure This);
static void* DbTStructureInstanciate (DbTStructure This);
static void DbTStructureInit (DbTStructure This, void* address);
static void DbTStructureErase (DbTStructure This, void* address);
static DbField DbTStructureFindField (DbTStructure This, char* name);
static DbTypeIterator DbTStructureGiveIterator (DbTStructure This, 
                                                IteratorMode mode, 
                                                void* address);
static int DbTStructureIsSame (DbTStructure This, DbTStructure other);
static DbType DbTStructureUsedStructure (DbTStructure This);

static DbTypeClassRec DbTStructureClass = {
  "DbTStructure",

  (DbRoutine) DbTStructureClear,
  (DbStringRoutine) DbTStructureName,
  (DbStringRoutine) NULL,
  (DbStringRoutine) NULL,
  (DbStringRoutine) NULL,
  (DbIntRoutine) DbTStructureSize,
  (DbIntRoutine) DbTStructureAlignment,
  (DbPtrRoutine) DbTStructureInstanciate,
  (DbInitRoutine) DbTStructureInit,
  (DbInitRoutine) DbTStructureErase,
  (DbIteratorRoutine) DbTStructureGiveIterator,
  (DbOutputRoutine) NULL,
  (DbCompareRoutine) DbTStructureIsSame,
  (DbUsedStructureRoutine) DbTStructureUsedStructure
};
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
static CLList  Structures = NULL;
/*---------------------------------------------------------------------*/

/*- procedure ---------------------------------------------------------*/
static void DbTStructureCleanup ()
/*---------------------------------------------------------------------*/
{
  CLListErase (Structures);
  GCReferenceClear (Structures);

  GCCleanup ();
}

/*- procedure ---------------------------------------------------------*/
DbType DbTStructureFind (char* name)
/*---------------------------------------------------------------------*/
{
  DbType result = NULL;
  CIterator it;

  if (Structures == NULL) return (NULL);

  it = CLListGiveIterator (Structures);
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbTStructure structure;

      structure = (DbTStructure) CIteratorItem (it);
      if (structure != NULL)
        {
          if (!strcmp (structure->name, name))
            {
              result = (DbType) structure;
              break;
            }
        }
    }
  CIteratorDelete (it);

  return (result);
}

/*- procedure ---------------------------------------------------------*/
int DbTStructureInstall (DbType This)
/*---------------------------------------------------------------------*/
/*                                                                     */
/*   Returns whether a true installation occured (1)                   */
/*  or (0) if the object was already installed.                        */
/*                                                                     */
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (0);
  if (!DbTypeIsStructure (This)) return (0);

  if (DbTStructureFind (DbTypeName (This)) != NULL) return (0);

  DbTypeStartup ();

  if (Structures == NULL)
    {
      GCReferenceAssign (Structures, CLListNew ());
    }

  CLListAdd (Structures, This);

  return (1);
}

/*- procedure ---------------------------------------------------------*/
int DbTStructureUninstall (DbType This)
/*---------------------------------------------------------------------*/
/*                                                                     */
/*   Returns whether a true un-installation occured (1)                */
/*  or (0) if the object was already uninstalled.                      */
/*                                                                     */
/*---------------------------------------------------------------------*/
{
  DbTStructure structure = NULL;
  CIterator it;

  if (This == NULL) return (0);
  if (!DbTypeIsStructure (This)) return (0);

  if (DbTStructureFind (DbTypeName (This)) == NULL) return (0);

  if (Structures == NULL) return (0);

  GCReferenceAssign (structure, This);

  CLListRemove (Structures, This);

  if (DbDebug ()) DbTypeCheck ();

  DbTypeUninstall (This);

  it = CLListGiveIterator (structure->fields);
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbField field = (DbField) CIteratorItem (it);
      DbFieldErase (field);

      GCReferenceClear (field->type);
    }
  CIteratorDelete (it);

  CLListErase  (structure->fields);
  GCReferenceClear (structure->fields);

  CHasherErase (structure->used_structures);
  GCReferenceClear (structure->used_structures);

  GCReferenceClear (structure);

  GCCleanup ();

  if (DbDebug ()) DbTypeCheck ();

  return (1);
}

/*- procedure ---------------------------------------------------------*/
void DbTStructureInstallFields (DbTStructure This, CLList fields)
/*---------------------------------------------------------------------*/
{
  CIterator it;
  int padding;

  GCReferenceAssignIn (This->fields, This, fields);

  it = CLListGiveIterator (fields);

  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbField field;
      DbType field_type;
      int offset;
      int alignment;
      
      field = (DbField) CIteratorItem (it);
      field_type = field->type;

      offset = DbTypeAlign (field_type, This->size);
      
      This->size += offset;
      field->offset = This->size;
      
      This->size += DbTypeSize (field_type);

      alignment = DbTypeAlignment (field_type);
      if (alignment > This->align) This->align = alignment;
    }
  
  CIteratorDelete (it);

  padding = DbTypeAlign ((DbType) This, This->size);
  This->size += padding;
}

/*- procedure ---------------------------------------------------------*/
DbType DbTStructureNew (char* name, CLList fields)
/*---------------------------------------------------------------------*/
{
  DbTStructure This;

  DbTypeStartup ();

  This = (DbTStructure) DbTStructureFind (name);
  if (This != NULL) 
    {
      if (fields == NULL) return ((DbType) This);

      if (This->fields == NULL)
        {
          DbTStructureInstallFields (This, fields);
          return ((DbType) This);
        }

      if (DbFieldListCompare (This->fields, fields)) return ((DbType) This);

        /*
          Same name as another structure found in the database but with
          different field list !

          So we create a new (uninstalled) object ???.
          */
    }

  This = (DbTStructure) malloc (sizeof (DbTStructureRec));
  DbTypeCount++;

  if (DbDebug ()) printf ("DbTStructureNew> type = %lx\n", This);


  This->base.classDef = &DbTStructureClass;

  This->name = (char*) malloc (strlen (name) + 1);
  DbStructureNameCount++;

  strcpy (This->name, name);

  This->fields          = NULL;
  This->size            = 0;
  This->align           = 0;
  This->used_structures = NULL;

  GCObjectNew (This, DbTypeGCClass);

  DbTStructureInstallFields (This, fields);

  return ((DbType) This);
}

/*- procedure ---------------------------------------------------------*/
static void DbTStructureClear (DbTStructure This)
/*---------------------------------------------------------------------*/
{
  if (DbDebug ()) printf ("DbTStructureClear> name= %s\n", This->name);

  if (This->name != NULL)
    {
      free (This->name);
      DbStructureNameCount--;

      This->name = NULL;
    }

  if (This->fields != NULL)
    {
      CLListErase (This->fields);
      GCReferenceClear (This->fields);
    }

  This->size = 0;
  This->align = 0;

  if (This->used_structures != NULL)
    {
      CHasherErase (This->used_structures);
      GCReferenceClear (This->used_structures);
    }
}

/*- procedure ---------------------------------------------------------*/
static char* DbTStructureName (DbTStructure This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (NULL);

  return (This->name);
}

/*- procedure ---------------------------------------------------------*/
static int DbTStructureSize (DbTStructure This)
/*---------------------------------------------------------------------*/
{
  return (This->size);
}

/*- procedure ---------------------------------------------------------*/
static int DbTStructureAlignment (DbTStructure This)
/*---------------------------------------------------------------------*/
{
  return (This->align);
}

/*- procedure ---------------------------------------------------------*/
static void* DbTStructureInstanciate (DbTStructure This)
/*---------------------------------------------------------------------*/
{
  void* object;

  object = (void*) malloc (This->size);
  memset (object, 0, This->size);

  DbTStructureInit (This, object);

  return (object);
}

/*- procedure ---------------------------------------------------------*/
static void DbTStructureInit (DbTStructure This, void* address)
/*---------------------------------------------------------------------*/
{
  CIterator it;
  char* ptr;

  it = CLListGiveIterator (This->fields);

  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbField field;
      DbType field_type;
      int offset;
      
      field = (DbField) CIteratorItem (it);
      field_type = field->type;
      offset     = field->offset;

      ptr = (char*) address + offset;

      DbTypeInit (field_type, ptr);
    }
  
  CIteratorDelete (it);
}

/*- procedure ---------------------------------------------------------*/
static void DbTStructureErase (DbTStructure This, void* address)
/*---------------------------------------------------------------------*/
{
  CIterator it;
  char* ptr;

  it = CLListGiveIterator (This->fields);

  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbField field;
      DbType field_type;
      int offset;
      
      field = (DbField) CIteratorItem (it);
      field_type = field->type;
      offset     = field->offset;

      ptr = (char*) address + offset;

      DbTypeErase (field_type, ptr);
    }
  
  CIteratorDelete (it);
}

/*- procedure ---------------------------------------------------------*/
static DbField DbTStructureFindField (DbTStructure This, char* name)
/*---------------------------------------------------------------------*/
{
  CIterator it;
  DbField result = NULL;
  
  if (This == NULL) return (NULL);

  it = CLListGiveIterator (This->fields);
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbField field;

      field = (DbField) CIteratorItem (it);

      if (!strcmp (name, field->name))
        {
          result = field;
          break;
        }
    }
  CIteratorDelete (it);

  return (result);
}

/*- procedure ---------------------------------------------------------*/
static DbTypeIterator DbTStructureGiveIterator (DbTStructure This, 
                                                IteratorMode mode, 
                                                void* address)
/*---------------------------------------------------------------------*/
{
  DbTypeIterator it;

  it =  DbTStructureIteratorNew (NULL, This, mode, address);

  return (it);
}

/*- procedure ---------------------------------------------------------*/
CIterator DbTStructureGiveFieldIterator (DbTStructure This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (NULL);

  return (CLListGiveIterator (This->fields));
}

/*---------------------------------------------------------------------*/
static int DbTStructureIsSame (DbTStructure This, DbTStructure other)
/*---------------------------------------------------------------------*/
{
  if (strcmp (This->name, other->name) != 0) return (0);

  return (DbFieldListCompare (This->fields, other->fields));
}

/*---------------------------------------------------------------------*/
static DbType DbTStructureUsedStructure (DbTStructure This)
/*---------------------------------------------------------------------*/
{
  return ((DbType) This);
}

/*---------------------------------------------------------------------*/
static int DbTStructureUseType (DbTStructure This, DbType type)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (0);

  if (This->used_structures == NULL) return (0);
  if (CHasherFind (This->used_structures, type) != NULL) return (1);

  return (0);
}

/*---------------------------------------------------------------------*/
int DbTStructureAddClient (DbTStructure This, DbType type)
/*---------------------------------------------------------------------*/
{
  if (type == NULL) return (0);

  if (This->used_structures == NULL)
    {
      GCReferenceAssignIn (This->used_structures, This, 
                           CPtrHasherNew (100, (CValuator) NULL));
    }
  else
    {
      if (CHasherFind (This->used_structures, type) != NULL) return (0);
    }

  CHasherAdd (This->used_structures, type);

  return (1);
}

/*---------------------------------------------------------------------*/
/*                                                                     */
/* What is questions.                                                  */
/*                                                                     */
/*---------------------------------------------------------------------*/

/*- procedure ---------------------------------------------------------*/
int DbTypeIsChar (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (0);

  return (This == DbTCharObject);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIsShort (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (0);

  return (This == DbTShortObject);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIsInt (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (0);

  return (This == DbTIntObject);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIsHexa (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (0);

  return (This == DbTHexaObject);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIsFloat (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (0);

  return (This == DbTFloatObject);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIsDouble (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (0);

  return (This == DbTDoubleObject);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIsString (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (0);

  return (This == DbTStringObject);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIsReference (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (0);

  return (This->classDef == &DbTReferenceClass);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIsVector (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (0);

  return (This->classDef == &DbTVectorClass);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIsMatrix (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (0);

  return (This->classDef == &DbTMatrixClass);
}

/*- procedure ---------------------------------------------------------*/
int DbTypeIsStructure (DbType This)
/*---------------------------------------------------------------------*/
{
  if (This == NULL) return (0);

  return (This->classDef == &DbTStructureClass);
}

/*- procedure ---------------------------------------------------------*/
static void DbTypeUninstall (DbType This)
/*---------------------------------------------------------------------*/
{
  CIterator it;

  if (This == NULL) return;

  it = CLListGiveIterator (Vectors);
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbTVector v = (DbTVector) CIteratorItem (it);

      if (DbTVectorUsedStructure (v) == This)
        {
          DbTVectorUninstall (v);
          break;
        }
    }
  CIteratorDelete (it);

  it = CLListGiveIterator (Matrices);
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbTMatrix m = (DbTMatrix) CIteratorItem (it);

      if (DbTMatrixUsedStructure (m) == This)
        {
          DbTMatrixUninstall (m);
          break;
        }
    }
  CIteratorDelete (it);

  it = CLListGiveIterator (References);
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbTReference r = (DbTReference) CIteratorItem (it);

      if (DbTReferenceUsedStructure (r) == This)
        {
          DbTReferenceUninstall (r);
          break;
        }
    }
  CIteratorDelete (it);

  it = CLListGiveIterator (Structures);
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbTStructure s = (DbTStructure) CIteratorItem (it);

      if (DbTStructureUseType (s, This))
        {
          DbTStructureUninstall ((DbType) s);
          break;
        }
    }
  CIteratorDelete (it);
}

